# ----------------------------------------------------------------------
#  COMPONENT: BackForth - slide two windows back/forth
#
#  This widget displays a window but can slide that out of the way
#  and slide another window into place.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itk
package require RapptureGUI

option add *BackForth.width 0 widgetDefault
option add *BackForth.height 0 widgetDefault
option add *BackForth.duration 300 widgetDefault

itcl::class BackForth {
    inherit Rappture::Notebook

    itk_option define -duration duration Duration 300

    constructor {args} {
        Rappture::Notebook::constructor
    } {
        set _tweener [Tweener ::#auto -duration 300]
        eval itk_initialize $args
    }

    public method current {args}
    public method position {page rel other}

    protected method _transition {from to val {dir "?"}}

    protected variable _current ""    ;# name of page being displayed
    protected variable _tweener ""    ;# manages animation for movements
    protected variable _trans         ;# x/y and 0-1 for page transitions
}

itk::usual BackForth {
    keep -cursor -background
}

# ----------------------------------------------------------------------
# USAGE: current ?<page>?
#
# With no args, this method returns the name of the current page.
# Otherwise, if <page> is not being displayed, it transitions from
# the old page to the new page.
# ----------------------------------------------------------------------
itcl::body BackForth::current {args} {
    if {[llength $args] == 0} {
        return $_current
    } else {
        set newpage [lindex $args 0]
        if {$newpage ne $_current} {
            if {$_current ne ""} {
                if {[info exists _trans($_current>$newpage)]} {
                    foreach {dir t0 t1} $_trans($_current>$newpage) break
                    $_tweener configure -from $t0 -to $t1 -command [itcl::code $this _transition $_current $newpage %v $dir] -finalize [itcl::code $this _transition $_current $newpage end]
                    $_tweener go -restart
                } elseif {[info exists _trans($newpage>$_current)]} {
                    foreach {dir t0 t1} $_trans($newpage>$_current) break
                    $_tweener configure -from $t1 -to $t0 -command [itcl::code $this _transition $newpage $_current %v $dir] -finalize [itcl::code $this _transition $_current $newpage end]
                    $_tweener go -restart
                } else {
                    # no transition path -- just show end frame
                    _transition $_current $newpage end
                }
            } else {
                # no starting frame -- just show end frame
                _transition $_current $newpage end
            }
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: position <page> -rightof|-leftof|-above|-below <otherPage>
#
# With no args, this method returns the name of the current page.
# Otherwise, if <page> is not being displayed, it transitions from
# the old page to the new page.
# ----------------------------------------------------------------------
itcl::body BackForth::position {page2 rel page1} {
    page $page2   ;# check for errors in page names
    page $page1

    switch -- $rel {
        -rightof { set _trans($page1>$page2) {x 0 1} }
        -leftof  { set _trans($page2>$page1) {x 0 1} }
        -above   { set _trans($page2>$page1) {y 0 1} }
        -below   { set _trans($page1>$page2) {y 0 1} }
        default {
            error "bad option \"$rel\": should be -rightof, -leftof, -above, -below"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: _transition <fromPage> <toPage> <val>|end ?<dir>?
#
# Handles the transition when switching between two panes of a window.
# The <val> is a number from 0-1.  If "end" is given instead, then
# the transition is completed and the final (landing) page is shown.
# ----------------------------------------------------------------------
itcl::body BackForth::_transition {from to val {dir "?"}} {
    if {$from ne ""} { set wfrom [page $from] }
    if {$to ne ""}   { set wto [page $to] }

    if {$val eq "end"} {
        if {$from ne ""} {
            place forget $wfrom
        }
        place $wto -x 0 -y 0 -relwidth 1 -relheight 1
        set _current $to
    } elseif {$to ne "" && $from ne ""} {
        switch -- $dir {
            x {
                set xwd [winfo width $itk_component(hull)]
                set x [expr {round((1-$val)*$xwd)}]
                place $wto -x $x -y 0 -relwidth 1 -relheight 1
                place $wfrom -x [expr {$x-$xwd}] -y 0 -relwidth 1 -relheight 1
            }
            y {
                set yht [winfo height $itk_component(hull)]
                set y [expr {round((1-$val)*$yht)}]
                place $wto -x 0 -y $y -relwidth 1 -relheight 1
                place $wfrom -x 0 -y [expr {$y-$yht}] -relwidth 1 -relheight 1
            }
            default { error "bad direction \"$dir\"" }
        }
    }
}

# ----------------------------------------------------------------------
# CONFIGURATION OPTION: -duration
# ----------------------------------------------------------------------
itcl::configbody BackForth::duration {
    if {![string is integer -strict $itk_option(-duration)]
          || $itk_option(-duration) <= 0} {
        error "bad value \"$itk_option(-duration)\": should be integer > 0"
    }
    $_tweener configure -duration $itk_option(-duration)
}
