# ----------------------------------------------------------------------
#  COMPONENT: citations for MD Showcase
#
#  This class is used to load/manage the citation data associated a
#  particular loaded showcase in the viewer.  Each citation is
#  represented in EndNote export format, since that's easy to parse.
#
#  Details about EndNote format are online here:
#  http://www.cardiff.ac.uk/insrv/educationandtraining/guides/endnote/endnote_codes.html
#
#  %0 = Reference type
#         Generic
#         Artwork
#         Audiovisual Material
#         Bill
#         Book
#         Book Section
#         Case
#         Chart or Table
#         Classical Work
#         Computer Program
#         Conference Paper
#         Conference Proceedings
#         Edited Book
#         Equation
#         Electronic Article
#         Electronic Book
#         Electronic Source
#         Figure
#         Film or Broadcast
#         Government Document
#         Hearing
#         Journal Article
#         Legal Rule/Regulation
#         Magazine Article
#         Manuscript
#         Map
#         Newspaper Article
#         Online Database
#         Online Multimedia
#         Patent
#         Personal Communication
#         Report
#         Statute
#         Thesis
#         Unpublished Work
#
#  %A = Author
#  %B = Secondary Title (of a Book or Conference Name)
#  %C = Place Published
#  %D = Year
#  %E = Editor /Secondary Author
#  %F = Label
#  %G = Language
#  %H = Translated Author
#  %I = Publisher
#  %J = Secondary Title (Journal Name)
#  %K = Keywords
#  %L = Call Number
#  %M = Accession Number
#  %N = Number (Issue)
#  %P = Pages
#  %Q = Translated Title
#  %R = Electronic Resource Number
#  %S = Tertiary Title
#  %T = Title
#  %U = URL
#  %V = Volume
#  %W = Database Provider
#  %X = Abstract
#  %Y = Tertiary Author
#  %Z = Notes
#  %0 = Reference Type
#  %1 = Custom 1
#  %2 = Custom 2
#  %3 = Custom 3
#  %4 = Custom 4
#  %6 = Number of Volumes
#  %7 = Edition
#  %8 = Date
#  %9 = Type of Work
#  %? = Subsidiary Author
#  %@ = ISBN/ISSN
#  %! = Short Title
#  %# = Custom 5
#  %$ = Custom 6
#  %] = Custom 7
#  %& = Section
#  %( = Original Publication
#  %) = Reprint Edition
#  %* = Reviewed Item
#  %+ = Author Address
#  %^ = Caption
#  %> = Link to PDF
#  %< = Research Notes
#  %[ = Access Date
#  %= = Last Modified Date
#  %~ = Name of Database
#
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itcl

itcl::class Citation {
    constructor {body} {
        foreach line [split $body \n] {
            set line [string trim $line]
            if {$line ne ""} {
                if {[regexp {^(%.)[ \t]+(.+)} $line match pcnt data]} {
                    lappend _fields($pcnt) $data
                } else {
                    error "bad citation format \"$line\": should be a field in EndNote export format"
                }
            }
        }
    }

    method get {args} {
        switch [llength $args] {
            0 {
                return [array get _fields]
            }
            1 {
                set val [lindex $args 0]
                if {[info exists _fields($val)]} {
                    return $_fields($val)
                }
                return ""
            }
            default {
                error "wrong # args: should be \"get\" or \"get %X\""
            }
        }
    }

    method produce {flavor} {
        switch -- $flavor {
            text {
                if {[info exists _fields(%B)]} {
                    set order {%A %T %B %E %P %I %D %@}
                } elseif {[info exists _fields(%J)]} {
                    set order {%A %T %J %V %P %D %1}
                } else {
                    set order {%A %T %D %U}
                }

                set result ""
                foreach f $order {
                    if {![info exists _fields($f)]} continue

                    set max [llength $_fields($f)]
                    set num 0
                    foreach val $_fields($f) {
                        switch $f {
                          %A - %E {
                            if {[string first , $val] >= 0} {
                                if {$num < $max-1} {
                                    append result "$val; "
                                } else {
                                    append result "$val, "
                                }
                            } else {
                                append result "$val, "
                            }
                            if {$f eq "%E" && $num == $max-1} {
                                append result "ed., "
                            }
                          }
                          %T {
                            set leading [string index $val 0]
                            if {$leading ne "\"" && $leading ne "'"} {
                                append result "\"$val\", "
                            } else {
                                append result "$val, "
                            }
                            append result ""
                          }
                          %B - %J {
                            append result "$val, "
                          }
                          %V {
                            set vstr $val
                            if {[info exists _fields(%N)]} {
                                set n [lindex $_fields(%N) 0]
                                append vstr "($n)"
                            }
                            append result "$vstr, "
                          }
                          %P {
                            set leading [string tolower [string index $val 0]]
                            if {$leading ne "p"} {
                                append result "pp. $val, "
                            } else {
                                append result "$val, "
                            }
                          }
                          %I {
                            set pub $val
                            if {[info exists _fields(%C)]} {
                                set loc [lindex $_fields(%C) 0]
                                append pub " ($loc)"
                            }
                            append result "$pub, "
                          }
                          %D {
                            if {[info exists _fields(%8)]} {
                                set extra [lindex $_fields(%8) 0]
                                set dstr "$extra $val"
                            } else {
                                set dstr $val
                            }
                            append result "$dstr, "
                          }
                          default {
                            append result "$val, "
                          }
                        }
                        incr num
                    }
                }
                # trim off trailing comma and whitespace
                regsub {, +$} $result "" result
                append result "."

                return $result
            }
            bibtex {
                return "no yet implemented"
            }
            endnote {
                set newline ""
                set result ""
                foreach f $reportTheseFields {
                    if {[info exists _fields($f)]} {
                        foreach elem $_fields($f) {
                            append result "$newline$f = $elem"
                            set newline "\n"
                        }
                    }
                }
                return $result
            }
            default {
                error "bad format \"$flavor\": should be bibtex, endnote, or text"
            }
        }
    }

    private variable _fields  ;# endnote data fields

    private common reportTheseFields
    lappend reportTheseFields %0 ;# Reference Type
    lappend reportTheseFields %A ;# Author
    lappend reportTheseFields %T ;# Title
    lappend reportTheseFields %B ;# Secondary Title (Book or Conference Name)
    lappend reportTheseFields %E ;# Editor /Secondary Author
    lappend reportTheseFields %J ;# Secondary Title (Journal Name)
    lappend reportTheseFields %V ;# Volume
    lappend reportTheseFields %N ;# Number (Issue)
    lappend reportTheseFields %P ;# Pages
    lappend reportTheseFields %I ;# Publisher
    lappend reportTheseFields %C ;# Place Published
    lappend reportTheseFields %D ;# Year
    lappend reportTheseFields %8 ;# Date
    lappend reportTheseFields %1 ;# Custom 1 (DOI)
    lappend reportTheseFields %K ;# Keywords
    lappend reportTheseFields %X ;# Abstract
    lappend reportTheseFields %@ ;# ISBN/ISSN
    lappend reportTheseFields %U ;# URL
}
