# ----------------------------------------------------------------------
#  COMPONENT: Controlbox - like a listbox with icons and controls
#
#  This widget is like a listbox, but also has icons and other details
#  for each item in the list.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itk

option add *Controlbox.width 3i widgetDefault
option add *Controlbox.padding 4 widgetDefault
option add *Controlbox.titleFont {Helvetica -12 bold} widgetDefault
option add *Controlbox.subtitleFont {Helvetica -12} widgetDefault
option add *Controlbox.foreground #999 widgetDefault
option add *Controlbox.entryBackground white widgetDefault
option add *Controlbox.entryForeground black widgetDefault
option add *Controlbox.selectedEntryBackground black widgetDefault
option add *Controlbox.selectedEntryForeground white widgetDefault
option add *Controlbox.arrowColor #666 widgetDefault
option add *Controlbox.hiliteBg #ffffcc widgetDefault
option add *Controlbox.hiliteFg black widgetDefault

itcl::class Controlbox {
    inherit itk::Widget

    itk_option define -titlefont titleFont Font ""
    itk_option define -subtitlefont subtitleFont Font ""
    itk_option define -foreground foreground Foreground ""
    itk_option define -entrybackground entryBackground Background ""
    itk_option define -entryforeground entryForeground Foreground ""
    itk_option define -selectedentrybackground selectedEntryBackground Background ""
    itk_option define -selectedentryforeground selectedEntryForeground Foreground ""
    itk_option define -hilitebg hiliteBg Foreground ""
    itk_option define -hilitefg hiliteFg Background ""
    itk_option define -arrowcolor arrowColor Foreground ""
    itk_option define -padding padding Padding 0

    itk_option define -selectcommand selectCommand SelectCommand ""

    constructor {args} { # defined below }

    public method insert {where args}
    public method itemconfigure {index args}
    public method delete {args}
    public method get {index args}
    public method select {index}
    public method size {}

    private method _entryForIndex {index}
    private method _redraw {args}
    private method _event {type x y}

    private variable _dispatcher ""  ;# dispatcher for !events
    private variable _items ""       ;# list of items in the controlbox
    private variable _data           ;# data for items on the _items list
    private variable _counter 0      ;# for generating unique entry IDs
    private variable _layout ""      ;# list of ypos/tid for all items
    private variable _hilite ""      ;# item with active hilite
    private variable _selected ""    ;# item with selected background
}

itk::usual Controlbox {
    keep -background -cursor
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Controlbox::constructor {args} {
    # create a dispatcher for events
    Rappture::dispatcher _dispatcher
    $_dispatcher register !redraw
    $_dispatcher dispatch $this !redraw [itcl::code $this _redraw]

    itk_component add list {
        canvas $itk_interior.list -highlightthickness 0 -height 1i
    } {
        keep -background -borderwidth -relief -width
    }
    pack $itk_component(list) -side left -expand yes -fill both

    bind $itk_component(list) <Motion> \
        [itcl::code $this _event motion %x %y]
    bind $itk_component(list) <ButtonPress> \
        [itcl::code $this _event click %x %y]
    bind $itk_component(list) <Leave> \
        [itcl::code $this _event motion -1 -1]

    bind $itk_component(list) <Configure> \
        [list $_dispatcher event -idle !redraw]

    eval itk_initialize $args
}

# ----------------------------------------------------------------------
# USAGE: insert <index> ?-tag <name> -option <value> ...?
#
# Inserts a new item into the listbox at the position <index>, which
# is an integer starting from 0 at the top of the list, or the keyword
# "end" for the very bottom of the list.  The remaining arguments are
# configuration options:
#
#   -tag .... gives the item a name that can be used in other ops
#   -text ... text displayed for the entry
#   -desc ... line below the -text heading
#   -data ... data associated with entry
# ----------------------------------------------------------------------
itcl::body Controlbox::insert {where args} {
    array set params $args
    if {[info exists params(-tag)]} {
        set tid $params(-tag)
        unset params(-tag)
    } else {
        set tid "node[incr _counter]"
    }

    if {[lsearch -exact $_items $tid] >= 0} {
        error "item \"$tid\" already exists -- tags must be unique"
    }
    set _items [linsert $_items $where $tid]
    set _data($tid-text) ""
    set _data($tid-desc) ""
    set _data($tid-data) ""

    eval itemconfigure $tid [array get params]

    # return the tag name for this new node
    return $tid
}

# ----------------------------------------------------------------------
# USAGE: itemconfigure <indexOrTag> -option <value> ...
#
# Used to modify an existing item.  The <indexOrTag> can be an integer
# index for the item or a -tag name given when the item was created.
# The remaining -option/value pairs are used to modify the icon/text
# and the data displayed in columns.
# ----------------------------------------------------------------------
itcl::body Controlbox::itemconfigure {index args} {
    set n [_entryForIndex $index]
    if {$n eq ""} {
        error "bad index \"$index\": should be integer or -tag name"
    }
    set tid [lindex $_items $n]

    # store the new data and then redraw
    foreach {key val} $args {
        if {[info exists _data($tid$key)]} {
            set _data($tid$key) $val

            # make sure we redraw the list at some point
            $_dispatcher event -idle !redraw
        } else {
            error "bad option \"$key\": should be -text, -desc, -data"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: delete ?<indexOrTag> <indexOrTag>...?
# USAGE: delete all
#
# Deletes one or more entries from this listbox.  With the "all"
# keyword, it deletes all entries in the listbox.  Otherwise, it
# deletes one or more entries as specified.
# ----------------------------------------------------------------------
itcl::body Controlbox::delete {args} {
    if {$args eq "all"} {
        set args $_items
    }
    foreach index $args {
        set n [_entryForIndex $index]
        if {$n ne ""} {
            set tid [lindex $_items $n]
            set _items [lreplace $_items $n $n]
            unset _data($tid-text)
            unset _data($tid-desc)
            unset _data($tid-data)

            # if this item was being hilited, then clear that
            if {$tid eq $_selected} {
                set _selected ""
            }

            # if this item was being hilited, then clear that
            if {$tid eq $_hilite} {
                set _hilite ""
            }

            # make sure we redraw the list at some point
            $_dispatcher event -idle !redraw
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: get <indexOrTag> ?-option -option...?
#
# Used to query information about an item.  With no extra args, it
# returns the tag with an entry, or "" if the entry is not recognized.
# If additional -option values are specified, then data values are
# returned for the requested options.
# ----------------------------------------------------------------------
itcl::body Controlbox::get {index args} {
    set n [_entryForIndex $index]
    if {$n eq ""} {
        return ""
    }
    if {[llength $args] == 0} {
        return [lindex $_items $n]
    }

    set vlist ""
    foreach option $args {
        if {[info exists _data($n$option)]} {
            lappend vlist $_data($n$option)
        } else {
            error "invalid option \"$option\""
        }
    }
    return $vlist
}

# ----------------------------------------------------------------------
# USAGE: select <indexOrTag>
#
# Used to select an item within the list.  If there is a -selectcommand,
# then it is invoked whenever the selection changes.
# ----------------------------------------------------------------------
itcl::body Controlbox::select {index} {
    set n [_entryForIndex $index]
    if {$n eq ""} {
        error "bad index \"$index\""
    }
    set tid [lindex $_items $n]

    if {$tid ne $_selected} {
        set c $itk_component(list)
        $c itemconfigure bg -outline $itk_option(-foreground) \
            -fill $itk_option(-entrybackground)
        $c itemconfigure text -fill $itk_option(-entryforeground)
        $c itemconfigure round -fill ""
        $c itemconfigure arrow -fill ""

        set bg $itk_option(-selectedentrybackground)
        set fg $itk_option(-selectedentryforeground)
        $c itemconfigure bg-$tid -outline "" -fill $bg
        $c itemconfigure round-$tid -fill $bg
        $c itemconfigure text-$tid -fill $fg
        $c itemconfigure arrow-$tid -fill $itk_option(-arrowcolor)

        $c raise bg-$tid
        $c raise text-$tid
        $c raise arrow-$tid

        if {$itk_option(-selectcommand) ne ""} {
            after idle $itk_option(-selectcommand) $_data($tid-data)
        }
    }
    set _selected $tid
    set _hilite ""
}

# ----------------------------------------------------------------------
# USAGE: size
#
# Returns the number of items in the listbox.
# ----------------------------------------------------------------------
itcl::body Controlbox::size {} {
    return [llength $_items]
}

# ----------------------------------------------------------------------
# USAGE: _entryForIndex <indexOrTag>
#
# Converts the given <indexOrTag> to a valid entry name.  If the
# <indexOrTag> is an integer, then it is interpreted as an index
# in the range 0 to the size of the list.  Otherwise, it is treated
# as a -tag name that may have been used when creating the entry.
#
# Returns an entry name or "" if the entry is not recognized.
# ----------------------------------------------------------------------
itcl::body Controlbox::_entryForIndex {index} {
    if {[string is integer -strict $index]} {
        return $index
    }
    if {$index eq "end"} {
        return [llength $_items]
    } elseif {$index eq "selected"} {
        return $_selected
    }
    set n [lsearch -exact $_items $index]
    if {$n >= 0} {
        return $n
    }
    return ""
}

# ----------------------------------------------------------------------
# USAGE: _redraw ?<eventArgs>...?
#
# Used internally to redraw all items in the list after it has changed.
# ----------------------------------------------------------------------
itcl::body Controlbox::_redraw {args} {
    set c $itk_component(list)
    set w [winfo width $c]
    $c delete all
    set _hilite ""

    set pad $itk_option(-padding)
    set tht [font metrics $itk_option(-titlefont) -linespace]
    set subtht [font metrics $itk_option(-subtitlefont) -linespace]
    set lineht [expr {$tht+$subtht+3*$pad}]
    set leftpad [expr {$lineht/2}]

    # run through the items and build the layout
    set _layout ""
    set ypos 0
    foreach tid $_items {
        set xpos [expr {$leftpad+$pad}]
        incr ypos $pad
        set y0 [expr {$ypos-$pad}]
        set y1 [expr {$ypos+$lineht+$pad}]
        set ymid [expr {$ypos+$lineht/2}]

        $c create oval 0 $y0 $lineht $y1 \
            -outline "" -fill "" -tags [list round round-$tid]
        $c create rectangle $leftpad $y0 $w [expr {$y1+1}] \
            -outline "" -fill "" -tags [list bg bg-$tid]

        set tsize [expr {$lineht/2-2*$pad}]
        $c create polygon \
            [expr {$w-$tsize-2*$pad}] [expr {$y0+2*$pad}] \
            [expr {$w-2*$pad}] $ymid \
            [expr {$w-$tsize-2*$pad}] [expr {$y1-2*$pad}] \
            -outline "" -fill "" -tags [list arrow arrow-$tid]

        set text $_data($tid-text)
        set desc $_data($tid-desc)
        set xtext [expr {$xpos+2*$pad}]

        if {$text ne "" && $desc ne ""} {
            $c create text $xtext [expr {$ymid-$pad/2}] -anchor sw \
                -text $text -font $itk_option(-titlefont) \
                -fill $itk_option(-foreground) -tags [list text text-$tid]
            $c create text $xtext [expr {$ymid+$pad/2}] -anchor nw \
                -text $desc -font $itk_option(-subtitlefont) \
                -fill $itk_option(-foreground) -tags [list text text-$tid]
        } elseif {$text ne ""} {
            $c create text $xtext $ymid -anchor w \
                -text $text -font $itk_option(-titlefont) \
                -fill $itk_option(-foreground) -tags [list text text-$tid]
        } elseif {$desc ne ""} {
            $c create text $text $ymid -anchor w \
                -text $desc -font $itk_option(-subtitlefont) \
                -fill $itk_option(-foreground) -tags [list text text-$tid]
        }

        if {$tid eq $_selected} {
            $c itemconfigure bg-$tid \
                -outline "" -fill $itk_option(-selectedentrybackground)
            $c itemconfigure round-$tid \
                -outline "" -fill $itk_option(-selectedentrybackground)
            $c itemconfigure text-$tid \
                -fill $itk_option(-selectedentryforeground)
            $c itemconfigure arrow-$tid \
                -fill $itk_option(-arrowcolor)
        } else {
            $c itemconfigure bg-$tid \
                -outline $itk_option(-foreground) \
                -fill $itk_option(-entrybackground)
            $c itemconfigure text-$tid \
                -fill $itk_option(-entryforeground)
        }

        set ypos [expr {$ypos + $lineht + $pad}]
        lappend _layout $ypos $tid
    }

    # fix the final size so the widget scrolls properly
    set x1 0; set y1 0
    foreach {x0 y0 x1 y1} [$c bbox all] break
    set x1 [expr {$x1+$pad}]
    set y1 [expr {$y1+$pad+1}]
    $c configure -scrollregion [list 0 0 $x1 $y1]

    # synthesize a motion event to pick up a new hilite
    set x [expr {[winfo pointerx $c]-[winfo rootx $c]}]
    set y [expr {[winfo pointery $c]-[winfo rooty $c]}]
    _event motion $x $y

    $itk_component(list) configure -height $y1
}

# ----------------------------------------------------------------------
# USAGE: _event <type> <x> <y>
#
# Used internally to handle the mouseover and click events on items
# within the listbox.  The <type> indicates what sort of event it
# is (motion or click), and the <x>,<y> is the coordinate relative
# to the listbox.  Uses the information in _layout to decide what
# item is being touched, and goes from there to handle the action.
# ----------------------------------------------------------------------
itcl::body Controlbox::_event {type x y} {
    set c $itk_component(list)

    set tid ""
    if {$y >= 0} {
        foreach {ypos item} $_layout {
            if {$y < $ypos} {
                set tid $item
                break
            }
        }
    }

    switch -- $type {
        motion {
            if {$tid ne $_hilite} {
                if {$_hilite ne ""} {
                    $c itemconfigure bg-$_hilite \
                        -fill $itk_option(-entrybackground)
                    $c itemconfigure text-$_hilite \
                        -fill $itk_option(-entryforeground)
                }

                if {$tid ne "" && $tid ne $_selected} {
                    # turn on hilite color
                    $c itemconfigure bg-$tid -fill $itk_option(-hilitebg)
                    $c itemconfigure text-$tid -fill $itk_option(-hilitefg)
                    set _hilite $tid
                }
            }
        }
        click {
            if {$tid ne ""} {
                select $tid
            }
        }
        default {
            error "don't know how to handle action \"$type\""
        }
    }
}

# ----------------------------------------------------------------------
#  CONFIGURATION OPTIONS: -hilitebg
# ----------------------------------------------------------------------
itcl::configbody Controlbox::hilitebg {
    if {$_hilite ne ""} {
        $itk_component(list) itemconfigure arrow-$_hilite \
            -fill $itk_option(-hilitebg)
    }
}

# ----------------------------------------------------------------------
#  CONFIGURATION OPTIONS: -entrybackground -entryforeground
#  CONFIGURATION OPTIONS: -selectedentrybackground
#  CONFIGURATION OPTIONS: -selectedentryforeground
# ----------------------------------------------------------------------
itcl::configbody Controlbox::entrybackground {
    $_dispatcher event -idle !redraw
}
itcl::configbody Controlbox::entryforeground {
    $_dispatcher event -idle !redraw
}
itcl::configbody Controlbox::selectedentrybackground {
    $_dispatcher event -idle !redraw
}
itcl::configbody Controlbox::selectedentryforeground {
    $_dispatcher event -idle !redraw
}
