# ----------------------------------------------------------------------
#  COMPONENT: Filebox - set of folders to explore a file system
#
#  BLT treeview, but meant to explore files in a file system.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2013  Purdue University
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itk
package require BLT

option add *Filebox.width 3i widgetDefault
option add *Filebox.height 1i widgetDefault
option add *Filebox.font {Helvetica -12} widgetDefault
option add *Filebox.background white widgetDefault
option add *Filebox.selectBackground #bbddff widgetDefault
option add *Filebox.selectForeground black widgetDefault

itcl::class Filebox {
    inherit itk::Widget

    itk_option define -selectbackground selectBackground Foreground ""
    itk_option define -selectforeground selectForeground Background ""
    itk_option define -fileroot fileRoot FileRoot ""

    constructor {args} { # defined below }
    destructor { # defined below }

    public method insert {args}
    public method delete {args}
    public method selection {option args}
    public method curselection {}
    public method xview {args}
    public method yview {args}

    private method _entryForIndex {index}
    private method _rebuild {args}
    private method _sort {args}

    private variable _dispatcher ""  ;# dispatcher for !events
    private variable _tree ""        ;# tree object within viewer
    private variable _file2index     ;# maps file name => index in tree
    private variable _index2file     ;# maps tree index => file name
    private variable _counter 0      ;# for generating unique entry IDs
}

itk::usual Filebox {
    keep -background -cursor -troughcolor
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Filebox::constructor {args} {
    itk_option add hull.width hull.height
    pack propagate $itk_component(hull) no

    # create a dispatcher for events
    Rappture::dispatcher _dispatcher
    $_dispatcher register !rebuild
    $_dispatcher dispatch $this !rebuild [itcl::code $this _rebuild]
    $_dispatcher register !sort
    $_dispatcher dispatch $this !sort [itcl::code $this _sort]

    # create a tree for file system info
    set _tree [blt::tree create]

    itk_component add treeview {
        blt::treeview $itk_interior.list -tree $_tree \
            -exportselection no -selectmode single \
            -activeicons "" -separator "\001" -autocreate no -flat no \
            -highlightthickness 0
    } {
        keep -background -foreground -font -borderwidth -relief
        keep -showtitles -hideroot
        keep -selectcommand
        keep -xscrollcommand -yscrollcommand
    }
    pack $itk_component(treeview) -expand yes -fill both

    # keep things sorted as nodes are opened and appear/disappear
    $itk_component(treeview) sort auto yes
    $itk_component(treeview) sort configure -decreasing 0 -column treeView

    eval itk_initialize $args
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Filebox::destructor {} {
    blt::tree destroy $_tree
}

# ----------------------------------------------------------------------
# USAGE: insert <fileOrDirectory>...
#
# Inserts one or more file names into the current file box.  Returns
# a list of nodes created by the insert.
# ----------------------------------------------------------------------
itcl::body Filebox::insert {args} {
    set rlist ""

    set rootlen [string length $itk_option(-fileroot)]
    if {$rootlen == 0} {
        error "can't insert until -fileroot is set"
    }

    foreach fname $args {
        if {$fname eq ""} {
            continue
        }
        if {![string equal -length $rootlen $itk_option(-fileroot) $fname]} {
            error "file \"$fname\" is not under -fileroot"
        }

        if {[info exists _file2index($fname)]} {
            set node $_file2index($fname)
        } else {
            # get the parent node -- create if necessary
            set tail [file tail $fname]
            set path [file dirname $fname]
            if {$path eq $itk_option(-fileroot)} {
                set parentNode 0
            } elseif {[info exists _file2index($path)]} {
                set parentNode $_file2index($path)
            } else {
                insert $path
                set parentNode $_file2index($path)
            }

            # insert a new node for this entry
            set node [$_tree insert $parentNode -label $tail]
            set _file2index($fname) $node
            set _index2file($node) $fname
            $_dispatcher event -idle !sort

            if {[file isdirectory $fname]} {
                foreach cfname [glob -nocomplain [file join $fname *]] {
                    insert $cfname
                }
            }
        }
        lappend rlist $node

        if {[file isdirectory $fname]} {
            set icons [Rappture::icon folder1]
        } else {
            switch -- [file extension $fname] {
                .dcd - .psf {
                    set icons [Rappture::icon molfile]
                }
                .sho {
                    set icons [Rappture::icon shofile]
                }
                default {
                    set icons [Rappture::icon file]
                }
            }
        }
        $itk_component(treeview) entry configure $node -icons $icons
    }
    return $rlist
}

# ----------------------------------------------------------------------
# USAGE: delete <fileOrDirectory>...
#
# Deletes one or more file names from the current file box.
# ----------------------------------------------------------------------
itcl::body Filebox::delete {args} {
    foreach fname $args {
        if {![info exists _file2index($fname)]} {
            error "file \"$fname\" is not in Filebox"
        }
        set node $_file2index($fname)

        $_tree delete $node
        unset _file2index($fname)
        unset _index2file($node)
    }
}

# ----------------------------------------------------------------------
# USAGE: selection clear <first> ?<last>?
# USAGE: selection clearall
# USAGE: selection includes <tagOrId>
# USAGE: selection set <first> ?<last>?
#
# Used to manipulate item select selection for the listbox.  Use
# "curselection" to query the current selection.
# ----------------------------------------------------------------------
itcl::body Filebox::selection {option args} {
    switch -- $option {
        clear - set {
            switch [llength $args] {
                1 {
                    set n [_entryForIndex [lindex $args 0]]
                    $itk_component(treeview) selection $option $n
                }
                2 {
                    set n1 [_entryForIndex [lindex $args 0]]
                    set n2 [_entryForIndex [lindex $args 1]]
                    $itk_component(treeview) selection $option $n1 $n2
                }
                default {
                    error "wrong # args: should be \"selection $option first ?last?\""
                }
            }
        }
        clearall {
            $itk_component(treeview) selection clearall
            $itk_component(treeview) selection anchor 0
        }
        includes {
            set n [_entryForIndex [lindex $args 0]]
            return [$itk_component(treeview) selection includes $n]
        }
        default {
            error "bad option \"$option\": should be clear, clearall, includes, set"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: curselection
#
# Returns indices for the current selection, or "" if there is no
# selection.
# ----------------------------------------------------------------------
itcl::body Filebox::curselection {} {
    set rlist ""
    foreach id [$itk_component(treeview) curselection] {
        if {[info exists _index2file($id)]} {
            lappend rlist $_index2file($id)
        } elseif {$id == 0} {
            lappend rlist $itk_option(-fileroot)
        }
    }
    return $rlist
}

# ----------------------------------------------------------------------
# USAGE: xview <option> <args>...
# USAGE: yview <option> <args>...
#
# Forwards scrolling commands on to the underlying widgets.
# ----------------------------------------------------------------------
itcl::body Filebox::xview {args} {
    eval $itk_component(treeview) xview $args
}
itcl::body Filebox::yview {args} {
    eval $itk_component(treeview) yview $args
}

# ----------------------------------------------------------------------
# USAGE: _entryForIndex <indexOrDir>
#
# Converts the given <indexOrDir> to a valid entry name.  If the
# <indexOrDir> is an integer, then it is interpreted as a tree index
# value.  Otherwise, it is treated as the name of a file/directory,
# which is translated to an integer index.
#
# Returns an entry name or "" if the entry is not recognized.
# ----------------------------------------------------------------------
itcl::body Filebox::_entryForIndex {index} {
    if {[string is integer -strict $index]} {
        return $index
    }
    if {[info exists _file2index($index)]} {
        return $_file2index($index)
    }
    if {$index == 0} {
        return $itk_option(-fileroot)
    }
    return ""
}

# ----------------------------------------------------------------------
# USAGE: _rebuild ?<eventArgs>...?
#
# Called whenever the -fileroot option changes to rebuild the contents
# of the widget.
# ----------------------------------------------------------------------
itcl::body Filebox::_rebuild {args} {
    $_tree delete 0
    catch {unset _file2index}
    catch {unset _index2file}

    $itk_component(treeview) entry configure 0 \
        -label "mdshowcase" -icons [Rappture::icon folder1]

    if {$itk_option(-fileroot) ne ""} {
        foreach fname [glob -nocomplain [file join $itk_option(-fileroot) *]] {
            insert $fname
        }
        $itk_component(treeview) open -recurse 0
    }
}

# ----------------------------------------------------------------------
# USAGE: _sort ?<eventArgs>...?
#
# Called after nodes are inserted to keep all entries properly sorted
# in alphabetical order.
# ----------------------------------------------------------------------
itcl::body Filebox::_sort {args} {
    #
    # HACK ALERT! the "sort auto" mode should work, but it seems like
    # we also need to call this to force it to happen
    #
    $itk_component(treeview) sort once -recurse 0
}

# ----------------------------------------------------------------------
#  CONFIGURATION OPTION: -fileroot
# ----------------------------------------------------------------------
itcl::configbody Filebox::fileroot {
    if {$itk_option(-fileroot) ne ""
          && ![file exists $itk_option(-fileroot)]} {
        error "bad value \"$itk_option(-fileroot)\": file/directory does not exist"
    }
    $_dispatcher event -idle !rebuild
}

# ----------------------------------------------------------------------
#  CONFIGURATION OPTIONS: -selectbackground / -selectforeground
# ----------------------------------------------------------------------
itcl::configbody Filebox::selectbackground {
    $itk_component(treeview) configure \
        -selectbackground $itk_option(-selectbackground) \
        -nofocusselectbackground $itk_option(-selectbackground)
}

itcl::configbody Filebox::selectforeground {
    $itk_component(treeview) configure \
        -selectforeground $itk_option(-selectforeground) \
        -nofocusselectforeground $itk_option(-selectforeground)
}
