# ----------------------------------------------------------------------
#  COMPONENT: FrameCache - most-recently used cache of VMD images
#
#  This class makes it easy to store images related to a trajectory
#  animation.  It stores images up to a certain limit, and discards
#  the least-recently used images.  It also looks for images that
#  should be requested and passes that information back to the caller
#  so the images can be requested.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Rappture
package require Itcl

# ----------------------------------------------------------------------
#  SAFE INTERPRETERS
# ----------------------------------------------------------------------
itcl::class FrameCache {
    # absolute maximum number of frames in cache (cannot be more than this)
    public variable max 100 {
        if {![string is integer -strict $max] || $max < 1} {
            error "bad value \"$max\": should be integer >= 1"
        }
        _pruneCache
    }

    # these can be set to indicate desired frame size
    # this is useful when the server sends smaller images during rotations
    public variable iwidth ""
    public variable iheight ""

    constructor {rootwin args} {
        if {![winfo exists $rootwin]} {
            error "bad value \"$rootwin\": should be widget"
        }
        set _rootwin $rootwin
        set _resizeimg [image create photo]
        eval configure $args
    }

    destructor {
        cleanup
        image delete $_resizeimg
    }

    # store data for an image in the cache
    # returns the image, which can then be displayed
    public method store {num data} {
        if {![info exists _frames($num)]} {
            set _frames($num) [_allocateFrame $num]
        }
        $_frames($num) configure -width 0 -height 0 -data $data
        catch {unset _waiting($num)}

        # if image is not the right size, then resize it
        if {$iwidth ne "" && $iheight ne ""} {
            set iw [image width $_frames($num)]
            set ih [image height $_frames($num)]
            if {$iwidth != $iw || $iheight != $ih} {
                $_resizeimg configure \
                    -width [image width $_frames($num)] \
                    -height [image height $_frames($num)]
                $_resizeimg copy $_frames($num)
                $_frames($num) configure -width $iwidth -height $iheight
                blt::winop resample $_resizeimg $_frames($num) sinc
            }
        }

        _usedInCache $num
        _pruneCache

        return $_frames($num)
    }

    # see if the image for a numbered frame exists
    public method exists {num} {
        return [info exists _frames($num)]
    }

    # see if we've already requested a particular frame
    public method waiting {args} {
        switch -- [llength $args] {
            0 {
                return [array size _waiting]
            }
            1 {
                set num [lindex $args 0]
                return [info exists _waiting($num)]
            }
            default {
                set option [lindex $args 0]
                set args [lrange $args 1 end]
                switch -- $option {
                    set {
                        foreach num $args {
                            set _waiting($num) 1
                        }
                    }
                    default {
                        error "bad option \"$option\": should be set"
                    }
                }
            }
        }
    }

    # get the image for a numbered frame
    # returns 1 if image is on-hand, with image in return variable
    # returns 0 if image should be requested
    public method get {num retVar} {
        upvar $retVar rval
        if {[info exists _frames($num)]} {
            set rval $_frames($num)
            _usedInCache $num
            return 1
        }
        return 0
    }

    # look ahead and see what frames are needed in direction of play
    # returns spec for frames that should be requested:
    #   frame # frame # range #:#:#
    public method lookAhead {num dir mode frame0 frame1 stepby} {
        if {$dir eq "pause"} {
            return ""
        }

        set critical ""
        set needed ""
        set nearterm [expr {$_size/4}]    ;# up to this many frames is critical
        set longterm [expr {3*$_size/4}]  ;# don't look out further than this

        for {set i 0} {$i < $longterm} {incr i} {
            # keep track of frames that we'll need soon
            if {![info exists _frames($num)] && ![info exists _waiting($num)]} {
                lappend needed $num
                if {$i < $nearterm} {
                    lappend critical $num
                }
            }

            # figure out the next frame
            switch -- $dir {
                play>> {
                    if {[incr num $stepby] > $frame1} {
                        switch -- $mode {
                            once {
                                # reached the end -- we're done
                                break
                            }
                            loop {
                                # looping -- continue at 0 frame
                                set num $frame0
                            }
                            rock {
                                # rocking -- change play direction
                                set dir "<<play"
                                incr num -$stepby
                            }
                            default {
                                error "bad mode \"$mode\": should be once, loop, or rock"
                            }
                        }
                    }
                }
                <<play {
                    if {[incr num -$stepby] < $frame0} {
                        switch -- $mode {
                            once {
                                # reached the end -- we're done
                                break
                            }
                            loop {
                                # looping -- continue at last frame
                                set num $frame1
                            }
                            rock {
                                # rocking -- change play direction
                                set dir "play>>"
                                incr num $stepby
                            }
                            default {
                                error "bad mode \"$mode\": should be once, loop, or rock"
                            }
                        }
                    }
                }
                default {
                    error "bad play direction \"$dir\": should be play>>, <<play, or pause"
                }
            }
            if {$num < $frame0 || $num > $frame1} break
        }

        # if there aren't many frames needed, then forget it for a while
        if {[llength $needed] < $_size/2} {
            if {[llength $critical] > 0} {
                set needed $critical
            } else {
                # not many needed, nothing critical -- skip it
                return ""
            }
        }

        # build a return result based on the list of frames needed
        set rval ""
        set r0 ""
        set r1 ""
        foreach num $needed {
            set _waiting($num) 1
            if {$r0 eq ""} {
                set r0 $num
                set r1 $num
            } elseif {[info exists considered($num)]} {
                if {$r0 < $r1} {
                    lappend rval range $r0:$stepby:$r1
                } elseif {$r0 > $r1} {
                    lappend rval range $r0:[expr {-$stepby}]:$r1
                } else {
                    lappend rval frame $r0
                }
                set r0 ""
                set r1 ""
                continue
            } elseif {abs($num-$r1) == $stepby} {
                set r1 $num
                set considered($num) 1
                continue
            } else {
                if {$r0 < $r1} {
                    lappend rval range $r0:$stepby:$r1
                } elseif {$r0 > $r1} {
                    lappend rval range $r0:[expr {-$stepby}]:$r1
                } else {
                    lappend rval frame $r0
                }
                set r0 $num
                set r1 $num
            }
        }

        if {$r0 ne ""} {
            if {$r0 < $r1} {
                lappend rval range $r0:$stepby:$r1
            } elseif {$r0 > $r1} {
                lappend rval range $r0:[expr {-$stepby}]:$r1
            } else {
                lappend rval frame $r0
            }
        }

        # grow/shrink the cache based on current memory size
        _adjustCache $frame0 $frame1

        return $rval
    }

    # clear the cache
    public method clear {} {
        # clean up all images and image holder widgets
        foreach num [array names _frames] {
            set imh $_frames($num)
            $imh configure -width 1 -height 1; $imh blank
            unset _frames($num)
        }
        catch {unset _waiting}
        set _recent ""
    }

    # clear the cache completely
    public method cleanup {} {
        # clean up all images and image holder widgets
        foreach num [array names _holders] {
            set label $_holders($num)
            image delete [$label cget -image]
            destroy $label
        }
        catch {unset _frames}
        catch {unset _waiting}
        catch {unset _holders}
        set _recent ""
        set _hcount 0
    }

    # make sure that an image exists to receive the frame
    public method _allocateFrame {num} {
        if {![info exists _holders($num)]} {
            # create a holder widget to speed colormap resolution
            set _holders($num) [label $_rootwin.holder[incr _hcount] \
                -image [image create photo]]
        }
        return [$_holders($num) cget -image]
    }

    # scan through all frames in the given range, and make sure that
    # there is enough memory to receive them.  If there's more than
    # enough, then adjust the cache size up.
    public method _adjustCache {frame0 frame1} {
        set headroom [expr {200*(1<<20)}]  ;# leave some free ram
        set framew [expr {($iwidth eq "") ? 300 : $iwidth}]
        set frameh [expr {($iheight eq "") ? 300 : $iheight}]

        if {[Rappture::sysinfo freeram] > $headroom && $_size < $max} {
            # run through the lookahead range and try to allocate image space
            for {set n $frame0} {$n < $frame1} {incr n} {
                if {[Rappture::sysinfo freeram] < $headroom} {
                    break
                }

                set imh [_allocateFrame $n]
                if {[image width $imh] <= 1 && [image height $imh] <= 1} {
                    # expand to allocate memory space for image
                    $imh configure -width $framew -height $frameh

                    if {[lsearch $_recent $n] < 0} {
                        # not using this right now -- bump up cache space
                        incr _size
                    }
                }
            }
        }
    }

    # move a frame to the front of the "recently used" list
    private method _usedInCache {num} {
        set i [lsearch $_recent $num]
        if {$i >= 0} {
            set _recent [lreplace $_recent $i $i]
        }
        set _recent [linsert $_recent 0 $num]
    }

    # free up any frames that are beyond the max in "recently used" list
    private method _pruneCache {} {
        while {[llength $_recent] > $_size} {
            set num [lindex $_recent end]
            set imh [$_holders($num) cget -image]
            $imh configure -width 1 -height 1; $imh blank
            catch {unset _frames($num)}

            set _recent [lreplace $_recent end end]
        }
    }

    private variable _rootwin ""       ;# holder widgets are children of this
    private variable _size 50          ;# current cache size
    private variable _frames           ;# maps frame num => image
    private variable _waiting          ;# maps frame num => lookAhead image
    private variable _recent ""        ;# numbers of frames recently used
    private variable _holders          ;# maps frame num => holder widget
    private variable _hcount 0         ;# unique counter for holder widgets
    private variable _resizeimg ""     ;# buffer for resizing images
}
