# ----------------------------------------------------------------------
#  COMPONENT: matrix4 - 4x4 rotation matrix, like the one in VMD
#
#  This class mimics the computation of rotation/scaling matrices
#  performed by VMD.  We compute much of this client side, and then
#  we can send the results to VMD via "set view_matrix" commands.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itcl

# ----------------------------------------------------------------------
#  SAFE INTERPRETERS
# ----------------------------------------------------------------------
itcl::class Matrix4 {
    constructor {{initialval ""}} {
        if {[catch {tmp get 0 0}] && [namespace tail $this] ne "tmp"} {
            # need one of these for internal calculations
            Matrix4 tmp
        }

        if {$initialval ne ""} {
            put $initialval
        } else {
            identity
        }
    }

    public method identity {} {
        put {{1 0 0 0} {0 1 0 0} {0 0 1 0} {0 0 0 1}}
    }

    public method constant {val} {
        for {set rnum 0} {$rnum < 4} {incr rnum} {
            for {set cnum 0} {$cnum < 4} {incr cnum} {
                set _vals($rnum,$cnum) $val
            }
        }
    }

    public method scale {factor} {
        tmp identity
        tmp put 0 0 $factor
        tmp put 1 1 $factor
        tmp put 2 2 $factor
        multmatrix tmp
    }

    public method translate {x y z} {
        tmp identity
        tmp put 3 0 $x
        tmp put 3 1 $y
        tmp put 3 2 $z
        multmatrix tmp
    }

    public method rotate {deg axis} {
        tmp identity
        set angle [expr {$deg*asin(1)/90.0}]
        switch -- $axis {
            x {
                tmp put 1 1 [expr {cos($angle)}]
                tmp put 2 2 [expr {cos($angle)}]
                tmp put 1 2 [expr {sin($angle)}]
                tmp put 2 1 [expr {-sin($angle)}]
            }
            y {
                tmp put 0 0 [expr {cos($angle)}]
                tmp put 2 2 [expr {cos($angle)}]
                tmp put 0 2 [expr {-sin($angle)}]
                tmp put 2 0 [expr {sin($angle)}]
            }
            z {
                tmp put 0 0 [expr {cos($angle)}]
                tmp put 1 1 [expr {cos($angle)}]
                tmp put 0 1 [expr {sin($angle)}]
                tmp put 1 0 [expr {-sin($angle)}]
            }
            default {
                error "bad axis \"$axis\": should be x, y, or z"
            }
        }
        multmatrix tmp
    }

    public method get {args} {
        switch -- [llength $args] {
            0 {
                # BEWARE: VMD writes matrices in column-major order
                set mtx ""
                for {set cnum 0} {$cnum < 4} {incr cnum} {
                    set row ""
                    for {set rnum 0} {$rnum < 4} {incr rnum} {
                        lappend row $_vals($rnum,$cnum)
                    }
                    lappend mtx $row
                }
                return $mtx
            }
            2 {
                set rnum [lindex $args 0]
                set cnum [lindex $args 1]
                return $_vals($rnum,$cnum)
            }
            default {
                error "wrong # args: should be \"get\" or \"get row col\""
            }
        }
    }

    public method put {args} {
        switch -- [llength $args] {
            1 {
                set mtx [lindex $args 0]
                if {[llength $mtx] != 4} {
                    error "bad matrix \"$mtx\": should be 4 rows with 4 elements each"
                }
                # BEWARE: VMD writes matrices in column-major order
                set cnum 0
                foreach col $mtx {
                    if {[llength $col] != 4} {
                        error "bad row \"$row\" in matrix: should have 4 elements"
                    }
                    set rnum 0
                    foreach val $col {
                        set _vals($rnum,$cnum) $val
                        incr rnum
                    }
                    incr cnum
                }
            }
            3 {
                set rnum [lindex $args 0]
                set cnum [lindex $args 1]
                set _vals($rnum,$cnum) [lindex $args 2]
            }
            default {
                error "wrong # args: should be \"put matrix\" or \"put row col val\""
            }
        }
    }

    public method multmatrix {obj} {
        for {set rnum 0} {$rnum < 4} {incr rnum} {
            for {set cnum 0} {$cnum < 4} {incr cnum} {
                set sum 0
                for {set n 0} {$n < 4} {incr n} {
                    set sum [expr {$sum+$_vals($rnum,$n)*[$obj get $n $cnum]}]
                }
                set new($rnum,$cnum) $sum
            }
        }

        # store the results back to the original matrix
        for {set rnum 0} {$rnum < 4} {incr rnum} {
            for {set cnum 0} {$cnum < 4} {incr cnum} {
                set _vals($rnum,$cnum) $new($rnum,$cnum)
            }
        }
    }

    private variable _vals     ;# values in the matrix
}
