# ----------------------------------------------------------------------
#  COMPONENT: parser - simple parser for Tcl-based command files
#
#  This class makes it easy to create a wide variety of Tcl-based
#  command parsers.  Use the "define" method to create the list
#  of commands recognized within a parser.  If the -strict option
#  is set to true, then any unrecognized commands are flagged as
#  errors; otherwise, unrecognized commands are ignored.  Each
#  body has access to a variable called "clientdata", which is
#  an array containing the data defined by the "clientdata" method
#  for the parser.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itcl

# ----------------------------------------------------------------------
#  SAFE INTERPRETERS
# ----------------------------------------------------------------------
itcl::class Rappture::Parser {
    public variable strict 1

    constructor {args} {
        set _nsname "[namespace current]::bodies[incr _counter]"
        namespace eval $_nsname {}
        set _parser [interp create -safe]
        foreach cmd [$_parser eval {info commands}] {
            $_parser hide $cmd
        }
        $_parser invokehidden proc unknown {args} {}
        $_parser expose error

        eval configure $args
    }

    destructor {
        interp delete $_parser
        namespace delete $_nsname
    }

    public method define {name arglist body} {
        set body "[list upvar #0 [itcl::scope _cdata] clientdata]\n$body"
        proc ${_nsname}::$name $arglist $body
        $_parser alias $name ${_nsname}::$name
        finalize
    }

    public method allow {args} {
        foreach cmd $args {
            $_parser expose $cmd
        }
    }

    public method clientdata {args} {
        switch -- [llength $args] {
            0 { return [array get _cdata] }
            1 { set key [lindex $args 0]; return $_cdata($key) }
            default { array set _cdata $args }
        }
    }

    public method load {script} {
        $_parser eval $script
    }

    public method commands {{pattern *}} {
        $_parser hide error
        $_parser hide unknown
        set cmds [$_parser invokehidden info commands $pattern]
        $_parser expose error
        $_parser expose unknown
        return $cmds
    }

    public method errorInfo {} {
        return [$_parser invokehidden set errorInfo]
    }

    private method finalize {} {
        $_parser hide error

        set cmds [lsort [$_parser invokehidden info commands]]
        set i [lsearch $cmds unknown]
        if {$i >= 0} {
            set cmds [lreplace $cmds $i $i]
        }
        $_parser expose error

        if {$strict} {
            $_parser invokehidden proc unknown {cmd args} [format {
                error "bad keyword \"$cmd\": should be %s"
            } [join $cmds {, }]]
        } else {
            $_parser invokehidden proc unknown {cmd args} { }
        }
    }

    private variable _parser ""  ;# safe interp within this parser
    private variable _nsname     ;# namespace containing command bodies
    private variable _cdata      ;# array of client data used within bodies
    private common _counter 0    ;# counter for unique _nsname names
}
