# ----------------------------------------------------------------------
#  COMPONENT: scenes for MD Showcase
#
#  This class is used to load/manage the scenes associated a particular
#  loaded showcase in the viewer.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require Itcl

itcl::class ShowcaseScene {
    common parser [Rappture::Parser ::#auto]
    common current ""   ;# stack of objects being built by parsers

    $parser define description {val} {
        set obj [lindex $::ShowcaseScene::current end]
        $obj configure -description $val
    }

    $parser define files {args} {
        set obj [lindex $::ShowcaseScene::current end]
        set flist ""

        # add unique file names onto the file list for this object
        set i 0
        while {$i < [llength $args]} {
            set fname [lindex $args $i]
            if {$fname eq "-name"} {
                incr i
                set fname [string trim [lindex $args $i]]
                if {[regexp {^[0-9]+$} $fname]} {
                    error "bad name \"$fname\" for file set: should start with alphabetical character"
                }
                lappend flist "@name:$fname"
            } else {
                if {[file pathtype $fname] ne "absolute"} {
                    error "bad file name \"$fname\": should be an absolute path"
                }
                set fname [file normalize $fname]
                if {[lsearch -exact $flist $fname] < 0} {
                    lappend flist $fname
                }
            }
            incr i
        }
        set oldflist [$obj cget -files]

        # if this didn't start with a name, give it a numeric one
        if {![regexp {^@name:} $flist]} {
            set count [regsub -all {@name:} $oldflist xx dummy]
            if {$count == 0} {
                set name "@name:top"
            } else {
                set name "@name:$count"
            }
            set flist [linsert $flist 0 $name]
        }

        $obj configure -files [concat $oldflist $flist]
    }

    $parser define script {body} {
        set obj [lindex $::ShowcaseScene::current end]
        $obj configure -vmdscript $body
    }

    $parser define view {args} {
        set obj [lindex $::ShowcaseScene::current end]
        if {[llength $args] < 1} {
            error "wrong # args: should be \"view ?-frames ##? ?play spec? body\""
        }

        $obj configure -viewscript [lindex $args end]

        foreach {key val} [lrange $args 0 end-1] {
            switch -- $key {
                -frame - -frames {
                    if {[regexp {^[0-9]+$} $val]} {
                        $obj configure -frames $val
                    } elseif {[regexp {^([0-9]+):([0-9]+)$} $val match v0 v1]} {
                        $obj configure -frames [list $v0 $v1]
                    } else {
                        error "bad frame specification \"$val\": should be a single integer or \"start:end\""
                    }
                }
                -play {
                    $obj configure -play $val
                }
                default {
                    error "bad option \"$key\": should be -frame/-frames or -play"
                }
            }
        }
    }

    $parser define settings {args} {
        set obj [lindex $::ShowcaseScene::current end]
        while {[llength $args] > 0} {
            set opt [lindex $args 0]
            if {[lsearch {-speed -smooth} $opt] >= 0} {
                if {[llength $args] > 1} {
                    set val [lindex $args 1]
                    if {![string is double -strict $val]} {
                        error "bad value \"$val\": should be double number"
                    }
                    $obj configure $opt $val
                    set args [lrange $args 2 end]
                } else {
                    error "missing value for option $opt"
                }
            } else {
                error "bad option \"$opt\": should be -speed, -smooth"
            }
        }
    }

    $parser define citation {body} {
        set obj [lindex $::ShowcaseScene::current end]
        $obj citations add $body
    }

    public variable name ""
    public variable description ""
    public variable files ""
    public variable speed ""
    public variable smooth ""
    public variable frames 0
    public variable play ""
    public variable vmdscript ""
    public variable viewscript ""

    constructor {args} {
        set info [lindex $args 0]
        if {$info ne "" && ![string match -* $info]} {
            # add this scene to the stack and parse the info
            lappend ::ShowcaseScene::current $this

            set status [catch { $::ShowcaseScene::parser load $info } result]

            set ::ShowcaseScene::current \
                [lrange $::ShowcaseScene::current 0 end-1]

            if {$status != 0} {
                error $result "\n    (while parsing scene data)"
            }
            set args [lrange $args 1 end]
        }

        # apply configuration options
        eval configure $args
    }

    destructor {
        foreach obj $_citations {
            itcl::delete object $obj
        }
    }

    public method citations {option args} {
        switch -- $option {
            add {
                if {[llength $args] != 1} {
                    error "wrong # args: should be \"citations add body\""
                }
                set body [lindex $args 0]
                lappend _citations [Citation ::#auto $body]
            }
            get {
                return $_citations
            }
            default {
                error "bad option \"$option\": should be add, get"
            }
        }
    }

    private variable _citations ""  ;# list of citation objects for scene
}

itcl::configbody ShowcaseScene::play {
    if {$play ne ""} {
        set op [lindex $play 0]
        set args [lrange $play 1 end]
        if {[lsearch {once loop rock pause paused} $op] < 0} {
            error "bad play spec \"$val\": should be \"once|loop|rock|paused ?forward|backward? ?from frame?"
        }

        set newvals(op) $op
        set newvals(from) ""
        set newvals(dir) "forward"

        set i 0
        while {$i < [llength $args]} {
            set key [lindex $args $i]
            switch -- $key {
                from {
                    set val [lindex $args [incr i]]
                    if {![string is int -strict $val] || $val < 0} {
                        error "bad frame number \"$val\": should be >= 0"
                    }
                    set newvals(from) $val
                }
                forward - forwards {
                    set newvals(dir) "forward"
                }
                backward - backwards {
                    set newvals(dir) "backward"
                }
                default {
                    error "bad play spec \"$play\": should be \"once|loop|rock|paused ?forward|backward? ?from frame?"
                }
            }
            incr i
        }

        # save the clean values back to the -play option
        set play [array get newvals]
    }
}
