# ----------------------------------------------------------------------
#  COMPONENT: VCR - play/pause controls like a VCR
#
#  This widget is used to control some sort of movie sequence via
#  standard VCR controls.
# ======================================================================
#  Copyright (c) 2013  Michael McLennan
# ======================================================================
package require BLT
package require Itk

option add *VCR.state normal widgetDefault
option add *VCR.font {Helvetica 10} widgetDefault
option add *VCR.presetColor #73b8e5 widgetDefault
option add *VCR.rangeColor yellow widgetDefault
option add *VCR.rangeLineColor black widgetDefault

blt::bitmap define VCRstripes {
#define stripes_width 9
#define stripes_height 9
static unsigned char stripes_bits[] = {
   0x92, 0x00, 0x49, 0x00, 0x24, 0x01, 0x92, 0x00, 0x49, 0x00, 0x24, 0x01,
   0x92, 0x00, 0x49, 0x00, 0x24, 0x01};
}

itcl::class VCR {
    inherit itk::Widget

    itk_option define -command command Command ""
    itk_option define -state state State ""
    itk_option define -font font Font ""
    itk_option define -range range Range ""
    itk_option define -rangecolor rangeColor RangeColor ""
    itk_option define -rangelinecolor rangeLineColor RangeLineColor ""
    itk_option define -presetcolor presetColor PresetColor ""
    itk_option define -max max Max 100

    constructor {args} { # defined below }

    method isplaying {args}
    method playmode {args}
    method progress {args}
    method speed {args}
    method smooth {args}

    private method _redraw {args}
    private method _fixFrame {args}
    private method _fixSpeed {args}
    private method _fixSmooth {args}
    private method _editor {option args}
    private method _event {type x y}
    private method _command {args}

    private variable _dispatcher ""    ;# dispatcher for !events
    private variable _playing 0        ;# 1=forward / 0=paused / -1=back
    private variable _playmode "once"  ;# one of: once, loop, rock
    private variable _progress 0       ;# position for play progress
    private variable _speed 0          ;# 0=normal / pos faster / neg slower
    private variable _speedset "--"    ;# if not --, then use as preset value
    private variable _smooth 0         ;# 0=normal / pos smooth / neg sharp
    private variable _smoothset "--"   ;# if not --, then use as preset value
    private variable _max 0            ;# max value for progress
    private variable _ymid 0           ;# y-coord for middle of widget
    private variable _yframe 0         ;# y-coord for bottom of frame num
    private variable _ycntls 0         ;# y-coord for controls beneath groove
    private variable _lcntls 0         ;# width of left-side controls
    private variable _rcntls 0         ;# width of right-side controls
    private variable _knob0 0          ;# x-coord for leftmost knob pos
    private variable _knob1 0          ;# x-coord for rightmost knob pos
    private variable _dragging ""      ;# part is being dragged

    # images for various VCR controls
    common _parts
    foreach name {fwd-on fwd-off fwd1 
        back-on back-off back1 pause-on plain
        groove-l groove-m groove-r groove-y knob
        speed-slow speed-fast smooth-lo smooth-hi
        run-once run-loop run-rock
    } {
        set imh [Rappture::icon vcr$name]
        if {$imh eq ""} { error "oops! can't find VCR part vcr$name" }
        set _parts($name) $imh
    }
}

itk::usual VCR {
    keep -background -cursor
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body VCR::constructor {args} {
    Rappture::dispatcher _dispatcher
    $_dispatcher register !redraw
    $_dispatcher dispatch $this !redraw [itcl::code $this _redraw]
    $_dispatcher register !frame
    $_dispatcher dispatch $this !frame [itcl::code $this _fixFrame]
    $_dispatcher register !speed
    $_dispatcher dispatch $this !speed [itcl::code $this _fixSpeed]
    $_dispatcher register !smooth
    $_dispatcher dispatch $this !smooth [itcl::code $this _fixSmooth]

    itk_component add area {
        canvas $itk_interior.area -borderwidth 0 -highlightthickness 0
    } {
        keep -background
    }
    pack $itk_component(area) -side left -expand yes -fill both

    bind $itk_component(area) <Configure> \
        [list $_dispatcher event -idle !redraw]

    # set a good minimum size for the widget
    set c $itk_component(area)
    set _lcntls [expr {[image width $_parts(fwd1)]
                        + [image width $_parts(fwd-off)] - 6}]
    set _rcntls [expr {[image width $_parts(back1)]
                        + [image width $_parts(back-off)] - 6}]

    set fn [option get $itk_component(hull) font Font]
    set numht [font metrics $fn -linespace]
    set grooveht [image height $_parts(groove-m)]
    set iconht [image height $_parts(speed-slow)]

    set ht [expr {$numht + 4 + $grooveht + 16 + $iconht}]
    set wd [expr {$_lcntls + [winfo pixels $c 1.5i] + $_rcntls}]
    $c configure -width $wd -height $ht

    set _ymid [expr {$numht+2+$grooveht/2}]
    set _yframe [expr {$_ymid-[image height $_parts(groove-m)]/2-1}]
    set _ycntls [expr {$_ymid+[image height $_parts(groove-m)]/2+5}]
    set _knob0 [expr {$_lcntls+3+[image width $_parts(groove-l)]}]
    set _knob1 [expr {$wd-$_rcntls-3-[image width $_parts(groove-r)]}]

    # create parts for the widget
    $c create image 2 $_ymid -anchor w -image $_parts(back1) -tags back1
    $c create image $_lcntls $_ymid -anchor e -image $_parts(back-off) -tags back

    $c create image [expr {$wd-2}] $_ymid -anchor e -image $_parts(fwd1) -tags fwd1
    $c create image [expr {$wd-$_rcntls}] $_ymid -anchor w -image $_parts(fwd-off) -tags fwd

    $c create image [expr {$_lcntls+6}] $_ymid -anchor w \
        -image [image create photo] -tags groove

    $c create rectangle -1 -1 -1 -1 -outline "" -stipple VCRstripes -tags range
    $c create line -1 -1 -1 -1 -fill "" -width 1 -tags range0
    $c create line -1 -1 -1 -1 -fill "" -width 1 -tags range1

    $c create image -20 -20 -anchor c -tags mode

    $c create image -20 -20 -anchor e -image $_parts(smooth-lo) -tags smooth0
    $c create image -20 -20 -anchor w -image $_parts(smooth-hi) -tags smooth1
    $c create rectangle -20 -20 -20 -20 -outline "" -fill gray -tags smoothbg
    $c create oval -20 -20 -20 -20 -width 2 -outline gray -fill gray -tags smoothset
    $c create line -20 -20 -20 -20 -width 1 -fill black -tags smoothpos

    $c create image -20 -20 -anchor e -image $_parts(speed-slow) -tags speed0
    $c create image -20 -20 -anchor w -image $_parts(speed-fast) -tags speed1
    $c create rectangle -20 -20 -20 -20 -outline "" -fill gray -tags speedbg
    $c create oval -20 -20 -20 -20 -width 2 -outline gray -fill gray -tags speedset
    $c create line -20 -20 -20 -20 -width 1 -fill black -tags speedpos

    $c create image -20 -20 -anchor center -image $_parts(knob) -tags knob

    $c create text -20 -20 -anchor s -text "0" -font $fn -tags frame

    $c create rectangle 0 0 $wd $ht \
        -outline "" -fill "" -stipple gray50 -tags screen

    bind $c <ButtonPress-1> [itcl::code $this _event click %x %y]
    bind $c <B1-Motion> [itcl::code $this _event drag %x %y]
    bind $c <ButtonRelease-1> [itcl::code $this _event release %x %y]

    # add tooltips for controls
    $c bind back1 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-back1 @%X,%Y"
    $c bind back1 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-back1 "Move backward one frame in the trajectory."

    $c bind back <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-back @%X,%Y"
    $c bind back <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-back "Play backward through the trajectory from the current point.  Click to start, and click again to stop."

    $c bind fwd <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-fwd @%X,%Y"
    $c bind fwd <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-fwd "Play forward through the trajectory from the current point.  Click to start, and click again to stop."

    $c bind fwd1 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-fwd1 @%X,%Y"
    $c bind fwd1 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-fwd1 "Move forward one frame in the trajectory."

    $c bind frame <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-frame @%X,%Y"
    $c bind frame <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-frame "Number of current frame, indicating the current point in the trajectory.  Click to edit the number and jump directly to another frame."

    $c bind mode <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-mode @%X,%Y"
    $c bind mode <Leave> "::Rappture::Tooltip::tooltip cancel"

    $c bind speed0 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-speed0 @%X,%Y"
    $c bind speed0 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-speed0 "Controls the speed of playback.  Drag the line toward this side to go slower."

    $c bind speed1 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-speed0 @%X,%Y"
    $c bind speed1 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-speed1 "Controls the speed of playback.  Drag the line toward this side to go faster."

    $c bind smooth0 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-smooth0 @%X,%Y"
    $c bind smooth0 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-smooth0 "Controls the smoothness of motion during playback.  Drag the line toward this side to be sharper."

    $c bind smooth1 <Enter> \
        "::Rappture::Tooltip::tooltip pending $c-smooth1 @%X,%Y"
    $c bind smooth1 <Leave> "::Rappture::Tooltip::tooltip cancel"
    Rappture::Tooltip::text $c-smooth1 "Controls the smoothness of motion during playback.  Drag the line toward this side to be smoother."

    bind $c <KeyPress> "::Rappture::Tooltip::tooltip cancel"

    # create an editor to type the frame value directly
    itk_component add editor {
        Rappture::Editor $itk_component(area).editor \
            -activatecommand [itcl::code $this _editor activate] \
            -validatecommand [itcl::code $this _editor validate] \
            -applycommand [itcl::code $this _editor apply]
    }

    eval itk_initialize $args

    progress 0

    # trigger callbacks that notify listeners of default values
    playmode once
    _command speed $_speed
    _command smooth $_smooth
}

# ----------------------------------------------------------------------
# USAGE: isplaying ?<value>?
#
# Used to set/get the "play" status of the video.  If zero, then the
# video is paused.  If positive, it is playing forward.  If negative,
# it is playing backward.
#
# If there's no argument, it returns the current status.
# ----------------------------------------------------------------------
itcl::body VCR::isplaying {args} {
    switch [llength $args] {
        0 {
            return $_playing
        }
        1 {
            if {$itk_option(-state) eq "disabled"} {
                return
            }
            set val [lindex $args 0]
            if {$val != $_playing && $_playing != 0} {
                # playing is suddenly reversed -- pause first
                set _playing 0
                $itk_component(area) itemconfigure fwd -image $_parts(fwd-off)
                $itk_component(area) itemconfigure back -image $_parts(back-off)
                $itk_component(area) itemconfigure fwd1 -image $_parts(fwd1)
                $itk_component(area) itemconfigure back1 -image $_parts(back1)
                _command pause
            }
            if {$val > 0} {
              if {$_playing != 1} {
                set _playing 1
                $itk_component(area) itemconfigure fwd -image $_parts(pause-on)
                $itk_component(area) itemconfigure back -image $_parts(back-off)
                $itk_component(area) itemconfigure fwd1 -image $_parts(plain)
                $itk_component(area) itemconfigure back1 -image $_parts(plain)
                _command play>>
              }
            } elseif {$val < 0} {
              if {$_playing != -1} {
                set _playing -1
                $itk_component(area) itemconfigure fwd -image $_parts(fwd-off)
                $itk_component(area) itemconfigure back -image $_parts(pause-on)
                $itk_component(area) itemconfigure fwd1 -image $_parts(plain)
                $itk_component(area) itemconfigure back1 -image $_parts(plain)
                _command <<play
              }
            } else {
              if {$_playing != 0} {
                set _playing 0
                $itk_component(area) itemconfigure fwd -image $_parts(fwd-off)
                $itk_component(area) itemconfigure back -image $_parts(back-off)
                $itk_component(area) itemconfigure fwd1 -image $_parts(fwd1)
                $itk_component(area) itemconfigure back1 -image $_parts(back1)
                _command pause
              }
            }
        }
        default {
            error "wrong # args: should be \"isplaying ?val?\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: playmode ?<newValue>?
#
# Used to set/get the play mode -- one of: once, loop, rock.  This
# affects what the video does when it reaches the end.  Mode "once"
# plays once and stops.  Mode "loop" starts over at the beginning.
# Mode "rock" goes back and forth.
#
# If there's no argument, it returns the mode.  Otherwise, it sets
# the current mode.
# ----------------------------------------------------------------------
itcl::body VCR::playmode {args} {
    switch [llength $args] {
        0 {
            return $_playmode
        }
        1 {
            set modes {once loop rock}
            set newval [lindex $args 0]
            if {$newval eq ">>"} {
                set i [lsearch $modes $_playmode]
                incr i
                if {$i >= [llength $modes]} {
                    set i 0
                }
                set newval [lindex $modes $i]
            } elseif {[lsearch $modes $newval] < 0} {
                error "bad value \"$newval\": should be one of: [join $modes {, }] or >>"
            }
            $itk_component(area) itemconfigure mode -image $_parts(run-$newval)

            if {$newval ne $_playmode} {
                set _playmode $newval
                _command mode $newval
            }

            # update the tooltip to report the mode
            set mesg "Controls how the sequence plays:  Play once, loop, or rock back and forth.  Currently "
            switch -- $_playmode {
                once { append mesg "playing once." }
                loop { append mesg "looping." }
                rock { append mesg "rocking back and forth." }
            }
            Rappture::Tooltip::text $itk_component(area)-mode $mesg
        }
        default {
            error "wrong # args: should be \"playmode ?val?\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: progress ?<position>? ?-nocommand?
#
# Used to set/get the progress through the video expressed as an integer
# from 0 to the -max value.  At 0, the knob is at the leftmost side
# (start), and at max, the knob is at the rightmost side (end).  If
# the -nocommand option is specified, then the -command option is not
# triggered for any notifications.  This is useful when the program
# is setting the position so that it shows up on the groove, but it
# does not need to react as though the user clicked and set the position.
#
# If there's no argument, it returns the current position.
# ----------------------------------------------------------------------
itcl::body VCR::progress {args} {
    # if there's a -nocommand option, take it out and set the flag
    set docmd 1
    set i [lsearch -exact $args "-nocommand"]
    if {$i >= 0} {
        set docmd 0
        set args [lreplace $args $i $i]
    }

    switch [llength $args] {
        0 {
            return $_progress
        }
        1 {
            if {$itk_option(-state) eq "disabled"} {
                return
            }
            set val [lindex $args 0]
            if {$val < 0} { set val 0 }
            if {$val > $_max} { set val $_max }

            if {$val ne $_progress} {
                set _progress $val

                if {$docmd} {
                    # invoke the callback to react to this new position
                    _command position $val
                }
            }

            $_dispatcher event -idle !frame
        }
        default {
            error "wrong # args: should be \"progress ?val?\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: speed ?-current|-present? ?<value>?
#
# Used to get/set the playback speed.  A value of "0" is normal (medium
# speed).  Negative values make the playback run slower, and positive
# values run faster.
#
# There are two values for speed: -current and -preset.  The -current
# value is in effect right now.  The -preset value is the suggested
# value for a given scene, or -- if there is no default.
#
# If there's no value specified, it returns the requested value.  If
# there's no flag, then it uses -current by default.
# ----------------------------------------------------------------------
itcl::body VCR::speed {args} {
    if {[llength $args] == 0} {
        set flag "-current"
    } else {
        set flag [lindex $args 0]
        if {[string is double -strict $flag]} {
            set flag "-current"
        } elseif {[lsearch {-current -preset} $flag] < 0} {
            error "bad flag $flag: should be -current or -preset"
        } else {
            # strip flag out of args
            set args [lrange $args 1 end]
        }
    }

    switch [llength $args] {
        0 {
            if {$flag eq "-preset"} {
                return $_speedset
            } else {
                return $_speed
            }
        }
        1 {
            if {$itk_option(-state) eq "disabled"} {
                return
            }

            set val [lindex $args 0]
            if {$flag eq "-preset" && $val eq "--"} {
                # this means no default
            } elseif {[string is double -strict $val]} {
                if {$val < -1} { set val -1 }
                if {$val > 1} { set val 1 }
            } else {
                error "bad speed \"$val\": should be double value"
            }

            switch -- $flag {
                -current {
                    if {$val != $_speed} {
                        set _speed $val
                        _command speed $val
                    }
                }
                -preset {
                    set _speedset $val
                }
            }

            # update the widget to show the speed change
            $_dispatcher event -idle !speed
        }
        default {
            error "wrong # args: should be \"speed ?which? ?val?\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: smooth ?-current|-preset? ?<value>?
#
# Used to get/set the playback smoothness.  A value of "0" is normal
# (medium smoothness).  Negative values smooth the animation for slow
# connections, and positive values sharpen the motion.
#
# There are two values for smoothness: -current and -preset.  The
# -current value is in effect right now.  The -preset value is the
# suggested value for a given scene, or -- if there is no default.
#
# If there's no value specified, it returns the requested value.  If
# there's no flag, then it uses -current by default.
# ----------------------------------------------------------------------
itcl::body VCR::smooth {args} {
    if {[llength $args] == 0} {
        set flag "-current"
    } else {
        set flag [lindex $args 0]
        if {[string is double -strict $flag]} {
            set flag "-current"
        } elseif {[lsearch {-current -preset} $flag] < 0} {
            error "bad flag $flag: should be -current or -preset"
        } else {
            # strip flag out of args
            set args [lrange $args 1 end]
        }
    }

    switch [llength $args] {
        0 {
            if {$flag eq "-preset"} {
                return $_smoothset
            } else {
                return $_smooth
            }
        }
        1 {
            if {$itk_option(-state) eq "disabled"} {
                return
            }

            set val [lindex $args 0]
            if {$flag eq "-preset" && $val eq "--"} {
                # this means no default
            } elseif {[string is double -strict $val]} {
                if {$val < -1} { set val -1 }
                if {$val > 1} { set val 1 }
            } else {
                error "bad speed \"$val\": should be double value"
            }

            switch -- $flag {
                -current {
                    if {$val != $_smooth} {
                        set _smooth $val
                        _command smooth $val
                    }
                }
                -preset {
                    set _smoothset $val
                }
            }

            # update the widget to show the smoothness change
            $_dispatcher event -idle !smooth
        }
        default {
            error "wrong # args: should be \"smooth ?which? ?val?\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: _redraw ?<eventArgs>...?
#
# Used internally to redraw items in the widget whenever the widget
# changes size.
# ----------------------------------------------------------------------
itcl::body VCR::_redraw {args} {
    set c $itk_component(area)
    set ht [winfo height $c]
    set wd [winfo width $c]
    $c coords screen 0 0 $wd $ht

    set _yframe [expr {$_ymid-[image height $_parts(groove-m)]/2-1}]
    set ybtm [expr {$_ymid+[image height $_parts(groove-m)]/2+5}]
    set _ycntls [expr {$ybtm + ($ht-$ybtm)/2}]
    set _knob1 [expr {$wd-$_rcntls-3-[image width $_parts(groove-r)]}]

    # reposition play controls according to the new height
    $c coords back1 2 $_ymid
    $c coords back $_lcntls $_ymid
    $c coords fwd1 [expr {$wd-2}] $_ymid
    $c coords fwd [expr {$wd-$_rcntls}] $_ymid

    $c coords groove [expr {$_knob0-[image width $_parts(groove-l)]}] $_ymid

    set xknob [expr {$_progress/double($_max)*($_knob1-$_knob0) + $_knob0}]
    $c coords knob $xknob [expr {$_ymid+1}]
    $c coords frame $xknob $_yframe

    # figure out the right-hand size of the knob
    set ght [image height $_parts(groove-m)]
    set glwd [image width $_parts(groove-l)]
    set grwd [image width $_parts(groove-r)]

    set gwd [expr {$_knob1-$_knob0}]
    if {$gwd < 1} {set gwd 1}

    # fix the layout of the background groove
    set imh [$c itemcget groove -image]
    $imh configure -width [expr {$glwd+$gwd+$grwd}] \
        -height [image height $_parts(groove-m)]

    $imh copy $_parts(groove-l) -to 0 0 $glwd $ght
    set x0 $glwd
    set xmax [expr {$glwd+$gwd}]
    set xchunk [image width $_parts(groove-m)]
    while {$gwd > 0} {
        if {$gwd > $xchunk} {
            set x1 [expr {$x0+$xchunk}]
        } else {
            set x1 $xmax
        }
        $imh copy $_parts(groove-m) -to $x0 0 $x1 $ght

        set gwd [expr {$gwd-($x1-$x0)}]
        set x0 $x1
    }
    $imh copy $_parts(groove-r) -to $xmax 0 [expr {$xmax+$grwd}] $ght

    # position the rectangle to highlight the specified -range
    if {[llength $itk_option(-range)] == 2 && $_max > 0} {
        foreach {r0 r1} $itk_option(-range) break

        set xr0 [expr {round(($_knob1-$_knob0)*$r0/double($_max)+$_knob0)}]
        if {$xr0 < $_knob0} {set xr0 $_knob0}
        if {$xr0 > $_knob1} {set xr0 $_knob1}

        set xr1 [expr {round(($_knob1-$_knob0)*$r1/double($_max)+$_knob0)}]
        if {$xr1 < $_knob0} {set xr1 $_knob0}
        if {$xr1 > $_knob1} {set xr1 $_knob1}

        $c coords range $xr0 [expr {$_ymid-8}] $xr1 [expr {$_ymid+8}]
        $c coords range0 $xr0 [expr {$_ymid-11}] $xr0 [expr {$_ymid+12}]
        $c coords range1 $xr1 [expr {$_ymid-11}] $xr1 [expr {$_ymid+12}]
    } else {
        $c coords range -1 -1 -1 -1
        $c coords range0 -1 -1 -1 -1
        $c coords range1 -1 -1 -1 -1
    }

    # reposition controls under the main bar
    set xmid [expr {($_knob0+$_knob1)/2}]
    $c coords mode $xmid $_ycntls

    set x0 [expr {7+[image width $_parts(speed-slow)]}]
    set x1 [expr {$xmid-[image width $_parts(run-once)]/2-25}]
    set x1 [expr {$x1-[image width $_parts(speed-fast)]}]
    $c coords speed0 [expr {$x0-2}] $_ycntls
    $c coords speed1 [expr {$x1+2}] $_ycntls
    $c coords speedbg $x0 [expr {$_ycntls-4}] $x1 [expr {$_ycntls+4}]

    _fixSpeed

    set x0 [expr {$xmid+[image width $_parts(run-once)]/2+25}]
    set x0 [expr {$x0+[image width $_parts(smooth-lo)]}]
    set x1 [expr {$wd-7-[image width $_parts(smooth-hi)]}]
    $c coords smooth0 [expr {$x0-2}] $_ycntls
    $c coords smooth1 [expr {$x1+2}] $_ycntls
    $c coords smoothbg $x0 [expr {$_ycntls-4}] $x1 [expr {$_ycntls+4}]

    _fixSmooth
}

# ----------------------------------------------------------------------
# USAGE: _fixFrame ?<eventArgs>...?
#
# Called whenever the frame position changes to update the position of
# the knob and the text shown for the frame number.
# ----------------------------------------------------------------------
itcl::body VCR::_fixFrame {args} {
    if {$_progress < 0} { set _progress 0 }
    if {$_progress > $_max} { set _progress $_max }

    set xknob [expr {$_progress/double($_max)*($_knob1-$_knob0) + $_knob0}]
    $itk_component(area) coords knob $xknob [expr {$_ymid+1}]
    $itk_component(area) coords frame $xknob $_yframe

    $itk_component(area) itemconfigure frame -text $_progress
}

# ----------------------------------------------------------------------
# USAGE: _fixSpeed ?<eventArgs>...?
#
# Called whenever the playback speed changes to update the controls
# showing the current value.
# ----------------------------------------------------------------------
itcl::body VCR::_fixSpeed {args} {
    set c $itk_component(area)
    foreach {x0 y0 x1 y1} [$c bbox speedbg] break

    # set the marker for the preset value
    if {$_speedset eq "--"} {
        $c coords speedset -20 -20 -20 -20
    } else {
        set xpos [expr {0.5*($_speedset+1)*($x1-$x0) + $x0}]
        set ymid [expr {0.5*($y0+$y1)}]
        $c coords speedset \
            [expr {$xpos-6}] [expr {$ymid-6}] \
            [expr {$xpos+6}] [expr {$ymid+6}]
    }

    # set the current value position
    set xpos [expr {0.5*($_speed+1)*($x1-$x0) + $x0}]
    $c coords speedpos $xpos [expr {$y0-2}] $xpos [expr {$y1+2}]
}

# ----------------------------------------------------------------------
# USAGE: _fixSmooth ?<eventArgs>...?
#
# Called whenever the playback speed changes to update the controls
# showing the current value.
# ----------------------------------------------------------------------
itcl::body VCR::_fixSmooth {args} {
    set c $itk_component(area)
    foreach {x0 y0 x1 y1} [$c bbox smoothbg] break

    # set the marker for the preset value
    if {$_speedset eq "--"} {
        $c coords smoothset -20 -20 -20 -20
    } else {
        set xpos [expr {0.5*($_smoothset+1)*($x1-$x0) + $x0}]
        set ymid [expr {0.5*($y0+$y1)}]
        $c coords smoothset \
            [expr {$xpos-6}] [expr {$ymid-6}] \
            [expr {$xpos+6}] [expr {$ymid+6}]
    }

    # set the current value position
    set xpos [expr {0.5*($_smooth+1)*($x1-$x0) + $x0}]
    $c coords smoothpos $xpos [expr {$y0-2}] $xpos [expr {$y1+2}]
}

# ----------------------------------------------------------------------
# USAGE: _editor popup
# USAGE: _editor activate
# USAGE: _editor validate <value>
# USAGE: _editor apply <value>
# USAGE: _editor menu <rootx> <rooty>
#
# Used internally to handle the various functions of the pop-up
# editor for the frame number.
# ----------------------------------------------------------------------
itcl::body VCR::_editor {option args} {
    if {$itk_option(-state) eq "disabled"} {
        return  ;# disabled? then bail out here!
    }
    switch -- $option {
        popup {
            $itk_component(editor) activate
        }
        activate {
            foreach {x0 y0 x1 y1} [$itk_component(area) bbox frame] break
            return [list text $_progress \
                x [expr {$x0+[winfo rootx $itk_component(area)]-1}] \
                y [expr {$y0+[winfo rooty $itk_component(area)]-1}] \
                w [expr {$x1-$x0+4}] \
                h [expr {$y1-$y0}]]
        }
        validate {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"_editor validate val\""
            }
            set val [lindex $args 0]

            if {![string is integer -strict $val] || $val < 0} {
                bell
                Rappture::Tooltip::cue $itk_component(editor) "Must be a positive integer"
                return 0
            }
        }
        apply {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"_editor apply val\""
            }
            set newval [lindex $args 0]
            progress $newval
        }
        default {
            error "bad option \"$option\": should be popup, activate, validate, apply"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: _event <type> <x> <y>
#
# Used internally to handle click events on VCR buttons.  The <type>
# indicates what sort of event it is (click/drag/release), and the
# <x>,<y> is the coordinate within the widget area.
# ----------------------------------------------------------------------
itcl::body VCR::_event {type x y} {
    Rappture::Tooltip::tooltip cancel
    if {$itk_option(-state) eq "disabled"} {
        return
    }
    set c $itk_component(area)

    if {$_dragging ne ""} {
        set tid $_dragging
    } else {
        set tid ""
        foreach id [$c find overlapping $x $y $x $y] {
            foreach tag [$c gettags $id] {
                if {[lsearch {fwd back fwd1 back1 knob groove mode frame speedbg speedpos speedset speed0 speed1 smoothbg smoothpos smoothset smooth0 smooth1} $tag] >= 0} {
                    set tid $tag
                    break
                }
            }
            if {$tid ne ""} {
                break
            }
        }

        switch -- $tid {
            speed0  { speed -1 ; return }
            speed1  { speed 1  ; return }
            smooth0 { smooth -1; return }
            smooth1 { smooth 1 ; return }
        }

        # look for any click near the frame number
        foreach {x0 y0 x1 y1} [$c bbox frame] break
        if {$x >= $x0-2 && $x <= $x1+2 && $y > $y0-2 && $y < $y1+2} {
            set tid frame
        }
    }

    switch -- $type {
        click {
            switch -- $tid {
                fwd - back {
                    # press in
                    foreach {x y} [$c coords $tid] break
                    $c coords $tid [expr {$x+2}] [expr {$y+2}]
                }
                fwd1 - back1 {
                    if {!$_playing} {
                        # press in
                        foreach {x y} [$c coords $tid] break
                        $c coords $tid [expr {$x+2}] [expr {$y+2}]
                    }
                }
                knob - groove {
                    set _dragging "knob"
                    _event drag $x $y
                }
                mode {
                    playmode >>
                }
                frame {
                    _editor popup
                }
                speedbg - speedpos - speedset {
                    set _dragging "speedbg"
                    _event drag $x $y
                }
                smoothbg - smoothpos - smoothset {
                    set _dragging "smoothbg"
                    _event drag $x $y
                }
            }
        }
        drag {
            switch -- $tid {
                knob - groove {
                    set dx [expr {$_knob1-$_knob0}]
                    set pos [expr {round(double($x-$_knob0)/$dx * $_max)}]
                    progress $pos
                }
                speedbg - speedpos {
                    foreach {x0 y0 x1 y1} [$c bbox speedbg] break
                    if {$x1 > $x0} {
                        if {$x < $x0} { set x $x0 }
                        if {$x > $x1} { set x $x1 }
                        set val [expr {2.0*($x-$x0)/($x1-$x0)-1}]

                        if {$_speedset ne "--" && abs($val-$_speedset) < 0.1} {
                            # near preset value? then snap to that
                            set val $_speedset
                        }

                        # don't call "speed" method until button release
                        set _speed $val
                        $_dispatcher event -idle !speed
                    }
                }
                smoothbg - smoothpos {
                    foreach {x0 y0 x1 y1} [$c bbox smoothbg] break
                    if {$x1 > $x0} {
                        if {$x < $x0} { set x $x0 }
                        if {$x > $x1} { set x $x1 }
                        set val [expr {2.0*($x-$x0)/($x1-$x0)-1}]

                        if {$_smoothset ne "--" && abs($val-$_smoothset) < 0.1} {
                            # near preset value? then snap to that
                            set val $_smoothset
                        }

                        # don't call "smooth" method until button release
                        set _smooth $val
                        $_dispatcher event -idle !smooth
                    }
                }
            }
        }
        release {
            switch -- $tid {
                fwd {
                    # pop back
                    foreach {x y} [$c coords $tid] break
                    $c coords $tid [expr {$x-2}] [expr {$y-2}]

                    # toggle the play status
                    if {$_playing > 0} {
                        isplaying 0
                    } else {
                        isplaying 1
                    }
                }
                back {
                    # pop back
                    foreach {x y} [$c coords $tid] break
                    $c coords $tid [expr {$x-2}] [expr {$y-2}]

                    # toggle the play status
                    if {$_playing < 0} {
                        isplaying 0
                    } else {
                        isplaying -1
                    }
                }
                fwd1 {
                    if {!$_playing} {
                        # pop back
                        foreach {x y} [$c coords $tid] break
                        $c coords $tid [expr {$x-2}] [expr {$y-2}]

                        # nudge forward
                        progress [expr {$_progress+1}]
                    }
                }
                back1 {
                    if {!$_playing} {
                        # pop back
                        foreach {x y} [$c coords $tid] break
                        $c coords $tid [expr {$x-2}] [expr {$y-2}]

                        # nudge backward
                        progress [expr {$_progress-1}]
                    }
                }
                knob - groove {
                    set _dragging ""
                    set dx [expr {$_knob1-$_knob0}]
                    set pos [expr {round(double($x-$_knob0)/$dx * $_max)}]
                    progress $pos
                }
                speedbg - speedpos {
                    # one last drag event
                    _event drag $x $y
                    set _dragging ""
                    _command speed [speed]
                }
                smoothbg - smoothpos {
                    # one last drag event
                    _event drag $x $y
                    set _dragging ""
                    _command smooth [smooth]
                }
            }
        }
        default {
            error "don't know how to handle action \"$type\""
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: _command <arg> <arg>...
#
# Used internally to handle click events on VCR buttons.  The <type>
# indicates what sort of event it is (click/drag/release), and the
# <x>,<y> is the coordinate within the widget area.
# ----------------------------------------------------------------------
itcl::body VCR::_command {args} {
    if {$itk_option(-command) ne ""} {
        set status [catch {uplevel #0 $itk_option(-command) $args} result]
        if {$status != 0} {
            bgerror "$result\n    (while handling -command callback for $itk_component(hull))"
        }
    }
}

# ----------------------------------------------------------------------
# OPTION: -state
# ----------------------------------------------------------------------
itcl::configbody VCR::state {
    switch -- $itk_option(-state) {
        normal {
            $itk_component(area) itemconfigure screen -fill ""
            $itk_component(area) lower screen
        }
        disabled {
            $itk_component(area) itemconfigure screen \
                -fill $itk_option(-background)
            $itk_component(area) raise screen
        }
        default {
            error "bad value \"$itk_option(-state)\": should be [join $allowed {, }]"
        }
    }
}

# ----------------------------------------------------------------------
# OPTION: -max
# ----------------------------------------------------------------------
itcl::configbody VCR::max {
    if {![string is integer -strict $itk_option(-max)]
          || $itk_option(-max) <= 0} {
        error "must be an integer value > 0"
    }
    set _max $itk_option(-max)

    if {$_progress > $_max} {set _progress $_max}

    $_dispatcher event -idle !redraw
    $_dispatcher event -idle !frame
}

# ----------------------------------------------------------------------
# OPTION: -range {frame frame}
# ----------------------------------------------------------------------
itcl::configbody VCR::range {
    if {$itk_option(-range) ne ""} {
        if {[llength $itk_option(-range)] != 2
              || ![string is integer -strict [lindex $itk_option(-range) 0]]
              || ![string is integer -strict [lindex $itk_option(-range) 1]]} {
            error "bad value \"$itk_option(-range)\": should be {int int}"
        }
    }
    $_dispatcher event -idle !redraw
}

# ----------------------------------------------------------------------
# OPTION: -font
# ----------------------------------------------------------------------
itcl::configbody VCR::font {
    $itk_component(area) itemconfigure frame -font $itk_option(-font)
    $_dispatcher event -idle !redraw
}

# ----------------------------------------------------------------------
# OPTION: -presetcolor, -rangecolor, -rangelinecolor
# ----------------------------------------------------------------------
itcl::configbody VCR::presetcolor {
    $itk_component(area) itemconfigure smoothset -fill $itk_option(-presetcolor)
    $itk_component(area) itemconfigure speedset -fill $itk_option(-presetcolor)
}

itcl::configbody VCR::rangecolor {
    $itk_component(area) itemconfigure range -fill $itk_option(-rangecolor)
}

itcl::configbody VCR::rangelinecolor {
    $itk_component(area) itemconfigure range0 -fill $itk_option(-rangelinecolor)
    $itk_component(area) itemconfigure range1 -fill $itk_option(-rangelinecolor)
}
