# ----------------------------------------------------------------------
#  COMPONENT: vmdvisviewer - view a molecule in 3D via VMD
#
#  This widget brings up a 3D representation of a molecule.  It uses
#  the VMD server on the rendering farm, so performance is hardware
#  accelerated.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2013  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Rappture
package require RapptureGUI

option add *VMDvisViewer.width 4i widgetDefault
option add *VMDvisViewer.height 4i widgetDefault
option add *VMDvisViewer.font {Helvetica -12} widgetDefault
option add *VMDvisViewer*loading*font {Helvetica -18 bold italic} widgetDefault

# must use this name -- plugs into Rappture::resources::load
proc VMDvisViewer_init_resources {} {
    Rappture::resources::register \
        vmdmds_server Rappture::VMDvisViewer::SetServerList
}

# ======================================================================
# Use this to store/retrieve the rotation matrices for an animation
# ======================================================================
itcl::class Rappture::VMDvisViewer::ViewSet {
    public method num {} {
        return [array size _views]
    }

    public method clear {} {
        catch {unset _views}
    }

    public method setview {which args} {
        if {![string is int $which]} {
            error "bad index \"$which\": should be integer"
        }
        set _views($which) $args
    }

    public method getview {which} {
        if {[info exists _views($which)]} {
            return $_views($which)
        }
        return ""
    }

    private variable _views  ;# maps frame number => view matrix (list form)
}

# ======================================================================
itcl::class Rappture::VMDvisViewer {
    inherit Rappture::VisViewer

    itk_option define -changecommand changeCommand ChangeCommand ""
    itk_option define -waitingcommand waitingCommand WaitingCommand ""

    constructor { args } {
        Rappture::VisViewer::constructor {}
    } {
        # defined below
    }
    destructor {
        # defined below
    }

    public method load {args}
    public method scene {option args}
    public method getimages {what spec framecb donecb}
    public method reset {{fromView "-initialview"}}
    public method rotate {option args}
    public method zoom {option args}
    public method pan {option x y}
    public method mouseMode {{mode ""}}
    public method play {option args}
    public method waiting {{state ""}}
    public method spinning {{state ""}}
    public method quality {{val ""}}

    private method DoResize {}
    private method ReceiveImage {size}
    private method ServerError {args}

    protected method Rebuild {}
    protected method BuildViews {name}
    protected method Mouse  {option x y func}

    protected method NewEpoch {}
    protected method MaxFrames {{max ""}}
    protected method SaveView {args}
    protected method BlankFrame {}
    protected method RefreshFrame {}
    protected method RequestFrames {}
    protected method NextFrame {num}
    protected method ShowFrame {num}
    protected method NextRotate {num max}
    protected method ShowRotate {num}
    protected method AutoSpinMatrix {num mtx}
    protected method AnimateWaiting {op {num ""}}
    protected method InvokeChangeCommand {args}

    public method Connect {}
    public method Disconnect {}

    public proc SetServerList {namelist} {
        Rappture::VisViewer::SetServerList "vmdmds" $namelist
    }

    private variable _data             ;# info about trajectory for vmd
    private variable _scenes           ;# maps scene name => config script
    private variable _scurr ""         ;# show this scene in _scenes
    private variable _camera           ;# properties of the current viewpoint
    private variable _mevent           ;# info used for mouse event operations
    private variable _mode ""          ;# mouse binding mode
    private variable _screenSize 1     ;# "good" measure of window size (pix)
    private variable _defview yes      ;# true => use defined views
    private variable _autospin         ;# maps frame num => rotation matrix
    private variable _waiting 0        ;# current state of "waiting" method
    private variable _spinning 0       ;# current state of "spinning" method
    private variable _waitinganim ""   ;# after id for "waiting" animation

    private variable _play             ;# params controlling play state/speed
    private variable _epoch 0          ;# epoch counter for frame requests
    private variable _cache ""         ;# cache for frames from server
    private variable _getimages        ;# params download images/animations

    private common func2cursor
    array set func2cursor {
        rotate exchange
        pan    fleur
        zoom   double_arrow
    }

    # ------------------------------------------------------------------
    # Create a parser that can evaluate many of the common VMD commands
    # and extract values--before we send them off to VMD for execution.
    # ------------------------------------------------------------------
    public proc ScanScript {script args} {
        Rappture::getopts args opts {
            flag how -reset default
            flag how -continue
        }

        # set up default values
        if {$opts(how) eq "-reset"} {
            array set scriptvals {
                backgroundgradient off
                color-Background black
                colordefs {blue #0000ff red #ff0000 gray #595959 orange #ff8000 yellow #ffff00 tan #808033 silver #999999 green #00ff00 white #ffffff pink #ff9999 cyan #40bfbf purple #a600a6 lime #80e566 mauve #e566b2 ochre #804d00 iceblue #8080bf black #000000 yellow2 #e0f705 yellow3 #8ce505 green2 #00e50a green3 #00e580 cyan2 #00e0ff cyan3 #00c2ff blue2 #0561ab blue3 #030aed violet #4500fa violet2 #7300e5 magenta #e500e5 magenta2 #ff00a8 red2 #fa003b red3 #cf0000 orange2 #e35900 orange3 #f5b800}
                rendermode Normal
            }
            set scriptvals(colorvals) $scriptvals(colordefs)
            set scriptvals(animateops) ""
            $scriptvals(rotatemtx) identity
            $scriptvals(scalemtx) identity
            $scriptvals(centermtx) identity
            $scriptvals(globalmtx) identity
        }

        if {[catch {$scriptparser load $script} result]} {
            puts stderr "VMD SCRIPT PARSING ERROR: $result"
        }
        return [array get scriptvals]
    }

    private common scriptvals
    set scriptvals(rotatemtx) [Matrix4 ::#auto]
    set scriptvals(scalemtx) [Matrix4 ::#auto]
    set scriptvals(centermtx) [Matrix4 ::#auto]
    set scriptvals(globalmtx) [Matrix4 ::#auto]

    private common scriptparser [Rappture::Parser ::#auto -strict 0]

    # allow some typical Tcl commands so that scripts function
    $scriptparser allow set lappend list if while switch foreach for

    # ------------------------------------------------------------------
    # VMD COMMAND: rotate -- changes rotation of the view
    # ------------------------------------------------------------------
    $scriptparser define rotate {args} {
        variable ::Rappture::VMDvisViewer::scriptvals

        if {[llength $args] == 3 || [llength $args] == 4} {
            set axis [lindex $args 0]
            set how [lindex $args 1]
            set angle [lindex $args 2]
            set step [lindex $args 3]

            if {[lsearch {x y z} $axis] >= 0} {

                # case "rotate axis by angle step" -- a little at a time
                if {$how eq "by" && $step ne "" && [string is double $step]} {
                    # rotate by this angle each step
                    set nsteps [expr {round($angle/double($step))}]
                    set angle $step
                    if {$nsteps-1 > 0} {
                        # keep rotating on subsequent frames
                        lappend scriptvals(animateops) [list \
                          [expr {$nsteps-1}] [list rotate $axis by $angle] \
                        ]
                    }
                }
                switch -- $how {
                    by {
                        $scriptvals(rotatemtx) rotate $angle $axis
                    }
                    to {
                        $scriptvals(rotatemtx) identity
                        $scriptvals(rotatemtx) rotate $angle $axis
                    }
                }
            }
        }
    }

    # ------------------------------------------------------------------
    # VMD COMMAND: scale -- changes zoom factor of the view
    # ------------------------------------------------------------------
    $scriptparser define scale {args} {
        variable ::Rappture::VMDvisViewer::scriptvals

        if {[llength $args] == 2} {
            set how [lindex $args 0]
            set factor [lindex $args 1]

            switch -- $how {
                by {
                    $scriptvals(scalemtx) scale $factor
                }
                to {
                    $scriptvals(scalemtx) identity
                    $scriptvals(scalemtx) scale $factor
                }
            }
        }
    }

    # ------------------------------------------------------------------
    # VMD COMMAND: translate -- shifts the center of the view
    # ------------------------------------------------------------------
    $scriptparser define translate {args} {
        variable ::Rappture::VMDvisViewer::scriptvals

        if {[llength $args] == 4} {
            set how [lindex $args 0]
            set dx [lindex $args 1]
            set dy [lindex $args 2]
            set dz [lindex $args 3]

            switch -- $how {
                by {
                    $scriptvals(centermtx) translate $dx $dy $dz
                }
                to {
                    $scriptvals(centermtx) identity
                    $scriptvals(centermtx) translate $dx $dy $dz
                }
            }
        }
    }

    # ------------------------------------------------------------------
    # VMD COMMAND: color -- redefines colors
    # ------------------------------------------------------------------
    $scriptparser define color {args} {
        variable ::Rappture::VMDvisViewer::scriptvals
        set op [lindex $args 0]
        if {$op eq "change" && [llength $args] >= 3} {
            if {[lindex $args 1] eq "rgb"} {
                set cname [lindex $args 2]
                switch -- [llength $args] {
                    3 {
                        # reset color to its default value
                        array set VMDcolors $scriptvals(colorvals)
                        array set VMDdefs $scriptvals(colordefs)
                        if {[info exists VMDdefs($cname)]} {
                            set VMDcolors($cname) $VMDdefs($cname)
                        }
                        set scriptvals(colorvals) [array get VMDcolors]
                    }
                    6 {
                        # set color to new RGB value
                        if {[catch {
                            foreach comp {r g b} val [lrange $args 3 5] {
                                set c($comp) [expr {round($val*255)}]
                                if {$c($comp) < 0} { set c($comp) 0 }
                                if {$c($comp) > 255} { set c($comp) 255 }
                            }
                           } result] == 0} {
                            array set VMDcolors $scriptvals(colorvals)
                            if {[info exists VMDcolors($cname)]} {
                                set VMDcolors($cname) [format \
                                    "#%02x%02x%02x" $c(r) $c(g) $c(b)]
                            }
                            set scriptvals(colorvals) [array get VMDcolors]
                        }
                    }
                }
            }
        } elseif {[llength $args] == 3} {
            # handle "color category name colorVal"
            set categ [lindex $args 0]
            if {$categ eq "Display"} {
                set name [lindex $args 1]
                set color [lindex $args 2]
                set scriptvals(color-$name) $color
            }
        }
    }

    # ------------------------------------------------------------------
    # VMD COMMAND: display -- changes background parameters
    # ------------------------------------------------------------------
    $scriptparser define display {args} {
        variable ::Rappture::VMDvisViewer::scriptvals
        if {[llength $args] == 2} {
            switch -- [lindex $args 0] {
                "backgroundgradient" {
                    set val [lindex $args 1]
                    if {[string is boolean -strict $val]} {
                        # save the status of the background gradient
                        set scriptvals(backgroundgradient) $val
                    }
                }
                "rendermode" {
                    set val [lindex $args 1]
                    if {[string equal -nocase "acrobat3d"]} {
                        set val "Normal"
                    }
                    set scriptvals(rendermode) $val
                }
            }
        }
    }
}

itk::usual VMDvisViewer {
    keep -background -cursor -font
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::constructor {args} {
    set _serverType "vmdmds"

    # Register events to the dispatcher.  Base class expects !rebuild
    # event to be registered.

    # Rebuild
    $_dispatcher register !rebuild
    $_dispatcher dispatch $this !rebuild "[itcl::code $this Rebuild]; list"

    # Resize event
    $_dispatcher register !resize
    $_dispatcher dispatch $this !resize "[itcl::code $this DoResize]; list"

    # Mouse Event
    $_dispatcher register !mevent
    $_dispatcher dispatch $this !mevent "[itcl::code $this _mevent]; list"

    # Populate the slave interpreter with commands to handle responses
    # from the visualization server.
    $_parser alias oops [itcl::code $this ServerError]
    $_parser alias image [itcl::code $this ReceiveImage]
    $_parser alias maxframes [itcl::code $this MaxFrames]
    $_parser alias saveview [itcl::code $this SaveView]
    $_parser alias rotatemtx [itcl::code $this AutoSpinMatrix]

    set _data(files) ""
    set _data(loaded) ""
    set _scenes(ALL) ""

    set _camera(zoom) 1
    set _camera(scale) 1
    set _camera(centerx) 0
    set _camera(centery) 0
    set _camera(centerz) 0
    set _camera(quality) "normal"

    array set _play {
        frame 0
        frame0 0
        frame1 0
        maxframes 0
        range ""
        state "pause"
        mode "once"
        afterId ""
        speed 0
        smooth 3
        delay 0
        frameby 1
        startfrom ""
        startop ""
        startdir "forward"
    }
    set _cache [FrameCache ::#auto $itk_interior -max 500]

    array set _getimages {
        framecallback ""
        donecallback ""
        remaining 0
    }

    # hide the usual sidebar controls
    pack forget $itk_component(main)

    itk_component add 3dview {
        label $itk_interior.view -image $_image(plot) \
            -highlightthickness 0 -borderwidth 0
    } {
        usual
        ignore -highlightthickness -borderwidth  -background
    }
    pack $itk_component(3dview) -expand yes -fill both

    # bindings for rotation mode
    bind rotate-$this <ButtonPress-1> \
        [itcl::code $this Mouse click %x %y rotate]
    bind rotate-$this <B1-Motion> \
        [itcl::code $this Mouse drag %x %y rotate]
    bind rotate-$this <ButtonRelease-1> \
        [itcl::code $this Mouse release %x %y rotate]

    # bindings for pan mode
    bind pan-$this <ButtonPress-1> \
        [itcl::code $this Mouse click %x %y pan]
    bind pan-$this <B1-Motion> \
        [itcl::code $this Mouse drag %x %y pan]
    bind pan-$this <ButtonRelease-1> \
        [itcl::code $this Mouse release %x %y pan]

    # bindings for zoom mode
    bind zoom-$this <ButtonPress-1> \
        [itcl::code $this Mouse click %x %y zoom]
    bind zoom-$this <B1-Motion> \
        [itcl::code $this Mouse drag %x %y zoom]
    bind zoom-$this <ButtonRelease-1> \
        [itcl::code $this Mouse release %x %y zoom]

    # allow these keypress bindings no matter what
    bind $itk_component(3dview) <KeyPress-Left> \
        [itcl::code $this pan by -1 0]
    bind $itk_component(3dview) <KeyPress-Right> \
        [itcl::code $this pan by 1 0]
    bind $itk_component(3dview) <KeyPress-Up> \
        [itcl::code $this pan by 0 -1]
    bind $itk_component(3dview) <KeyPress-Down> \
        [itcl::code $this pan by 0 1]
    bind $itk_component(3dview) <Shift-KeyPress-Left> \
        [itcl::code $this pan by -5 0]
    bind $itk_component(3dview) <Shift-KeyPress-Right> \
        [itcl::code $this pan by 5 0]
    bind $itk_component(3dview) <Shift-KeyPress-Up> \
        [itcl::code $this pan by 0 -5]
    bind $itk_component(3dview) <Shift-KeyPress-Down> \
        [itcl::code $this pan by 0 5]
    bind $itk_component(3dview) <KeyPress-Prior> \
        [itcl::code $this zoom out 2]
    bind $itk_component(3dview) <KeyPress-Next> \
        [itcl::code $this zoom in 2]

    bind $itk_component(3dview) <Enter> "focus $itk_component(3dview)"

    if {[string equal "x11" [tk windowingsystem]]} {
        bind $itk_component(3dview) <4> [itcl::code $this zoom out 1.1]
        bind $itk_component(3dview) <5> [itcl::code $this zoom in 1.1]
    }

    bind $itk_component(3dview) <Configure> [itcl::code $this DoResize]

    # create a pop-up window to show status while things are loading
    toplevel $itk_component(hull).loading -borderwidth 1 -relief solid
    wm overrideredirect $itk_component(hull).loading yes
    wm withdraw $itk_component(hull).loading
    label $itk_component(hull).loading.text -text "Loading..."
    pack $itk_component(hull).loading.text -padx 20 -pady {20 2}
    label $itk_component(hull).loading.spinner -image [Rappture::icon roller1]
    pack $itk_component(hull).loading.spinner -padx 4 -pady {2 20}

    # create a mini loading window to loading during animations
    toplevel $itk_component(hull).play -borderwidth 0
    wm overrideredirect $itk_component(hull).play yes
    wm withdraw $itk_component(hull).play
    label $itk_component(hull).play.spinner -image [Rappture::icon roller1]
    pack $itk_component(hull).play.spinner -padx 4 -pady 4

    eval itk_initialize $args

    reset
    Connect
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::destructor {} {
    VisViewer::Disconnect
    image delete $_image(plot)
    itcl::delete object $_cache
}

# ----------------------------------------------------------------------
# USAGE: load ?<trajectoryFile> <psfFile>...?
#
# Clients use this to set or query the trajectory files to scene.
# With no args, it returns the current list of files for the scene.
# Otherwise, it sets the require files to the given list of files.
# Returns 1 if these are new files--in which case a new scene will
# be loaded on the server.  Returns 0 if these files are already
# loaded and the load is unnecessary.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::load {args} {
    if {$args eq ""} {
        return $_data(files)
    }

    if {$args ne $_data(files)} {
        set _data(files) $args

        if { [IsConnected] } {
            $_dispatcher event -idle !rebuild
        } else {
            Connect
        }
        return 1  ;# files will be loaded
    }
    return 0  ;# files are already loaded
}

# ----------------------------------------------------------------------
# USAGE: scene define -name <name> -files <fileList> \
#                    -vmdscript <script> -frames <nums> -play <spec> \
#                    -viewscript <script>
# USAGE: scene show <name>
# USAGE: scene clear
# USAGE: scene forget ?<name> <name>...?
# USAGE: scene vmdinfo <name>
# USAGE: scene initialview <name>
# USAGE: scene lastview <name>|-prep
#
# Used to define and manipulate scenes of the trajectory information
# loaded previously by the "load" command.  The "define" operation
# defines the script that loads a scene called <name>.  The "show"
# operation executes that script to show the scene.  The "clear"
# operation clears the current scene (usually in preparation for 
# showing another scene).  The "forget" operation erases one or more
# scene definitions; if no names are specified, then all scenes are
# forgotten.  The "vmdinfo" operation queries any values that were
# extracted from the -vmdscript evaluation.  The "initialview" operation
# queries the rotation/scale/center/global matrices and the starting
# frame for the initial view of the scene.  The "lastview" operation
# queries the same thing for the last (current) view and returns an
# empty string if the last view is the same as the default.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::scene {option args} {
    switch -- $option {
        define {
            Rappture::getopts args opts {
                value -name ""
                value -files ""
                value -vmdscript ""
                value -frames 0
                value -play ""
                value -viewscript ""
            }
            if {[llength $args] != 0} {
                error "wrong # args: should be \"scene define -opt val ...\""
            }

            foreach key {-name -files -vmdscript} {
                if {$opts($key) eq ""} {
                    error "bad scene definition: $args (mising $key)"
                }
            }
            if {[lsearch -exact $_scenes(ALL) $opts(-name)] >= 0} {
                error "scene \"$opts(-name)\" already defined"
            }

            set name $opts(-name)
            lappend _scenes(ALL) $name
            set _scenes($name-files) $opts(-files)
            set _scenes($name-vmdscript) $opts(-vmdscript)
            set _scenes($name-frames) $opts(-frames)
            set _scenes($name-play) $opts(-play)
            set _scenes($name-viewscript) $opts(-viewscript)
            set _scenes($name-show) "-initialview"
            set _scenes($name-initialview) ""
            set _scenes($name-lastview) ""
            set _scenes($name-initialviewframe) ""
            set _scenes($name-lastviewframe) ""
            set _scenes($name-views) [ViewSet ::#auto]

            # set the $mol() var used in the vmd script
            $scriptparser load [list catch {unset mol}]
            $scriptparser load [list set mol(top) 0]
            set slot 0
            foreach fname $_scenes($name-files) {
                if {[regexp {^@name:(.+)} $fname match id]} {
                    $scriptparser load [list set mol($id) 0]
                    $scriptparser load [list set mol($slot) 0]
                    incr slot
                }
            }

            # scan the overall vmd script and capture view settings
            set _scenes($name-vmdinfo) [ScanScript $opts(-vmdscript)]

            $_dispatcher event -idle !rebuild
        }
        show {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"scene show name\""
            }
            set name [lindex $args 0]
            if {[lsearch -exact $_scenes(ALL) $name] < 0} {
                error "bad scene name \"$name\": should be one of [join $_scenes(ALL) {, }]"
            }

            # save characteristics of the current view
            if {$_scurr ne ""} {
                if {$_defview} {
                    set _scenes($_scurr-show) -initialview
                } else {
                    set _scenes($_scurr-show) -lastview
                }
                set _scenes($_scurr-lastviewframe) $_play(frame)
                SendBytes "tellme {saveview lastview %v} getview\n"
            }

            # install the next view...
            set _scurr $name
            BlankFrame
            waiting on

            # set the play characteristics
            array set _play { startop "" startfrom 0 startdir forward range "" }
            if {$_scenes($name-show) eq "-initialview"} {
                # initial view may play according to view definitions
                foreach {key val} $_scenes($name-play) {
                    set _play(start$key) $val
                }
                if {[llength $_scenes($name-frames)] == 2} {
                    set _play(range) $_scenes($name-frames)
                }

                set firstframe $_play(startfrom)
                if {$firstframe eq ""} {
                    set firstframe [lindex $_play(range) 0]
                }
                set _scenes($_scurr-initialviewframe) $firstframe
            } else {
                # for any other view, just go to that frame
                set show $_scenes($name-show)
                set firstframe $_scenes($name${show}frame)
            }

            if {$firstframe ne ""} {
                set _play(startfrom) $firstframe
            }

            if {![eval load $_scenes($name-files)]} {
                # files are already loaded -- just change the scene
                BuildViews $name

                NewEpoch
                SendBytes "scene show [list $name] -after {saveview initialview %v}\n"

                # reset the view and request a new frame
                reset $_scenes($name-show)

                # if the -play option is set, start playing
                if {[lsearch {once loop rock} $_play(startop)] >= 0} {
                    InvokeChangeCommand playmode $_play(startop) $_play(startdir)
                }
            }
        }
        clear {
            NewEpoch
            SendBytes "scene clear\n"
            RefreshFrame
        }
        vmdinfo {
            if {[llength $args] < 1} {
                error "wrong # args: should be \"scene vmdinfo name ?option option ...?\""
            }
            set name [lindex $args 0]

            if {[info exists _scenes($name-vmdinfo)]} {
                array set data $_scenes($name-vmdinfo)
            }

            switch -- [llength $args] {
                1 {
                    # no extra args? return all data
                    return [array get data]
                }
                2 {
                    # one extra arg? then return value for that
                    set opt [lindex $args 1]
                    if {[info exists data($opt)]} {
                        return $data($opt)
                    }
                    return ""
                }
                default {
                    # many extra args? return a list of values
                    set rlist ""
                    foreach opt $args {
                        if {[info exists data($opt)]} {
                            lappend rlist $data($opt)
                        } else {
                            lappend rlist ""
                        }
                    }
                    return $rlist
                }
            }

            foreach opt $args {
            }
        }
        initialview {
            set name [lindex $args 0]
            if {[info exists _scenes($name-initialview)]} {
                set info $_scenes($name-initialview)
                lappend info -frame $_scenes($name-initialviewframe)
                return $info
            }
            return ""
        }
        lastview {
            if {$args eq "-prep"} {
                if {$_defview} {
                    set _scenes($_scurr-show) -initialview
                } else {
                    set _scenes($_scurr-show) -lastview
                }
                set _scenes($_scurr-lastviewframe) $_play(frame)
                SendBytes "tellme {saveview lastview %v} getview\n"
                return
            }

            set name [lindex $args 0]
            if {[info exists _scenes($name-show)]
                  && $_scenes($name-show) eq "-lastview"} {
                set info $_scenes($name-lastview)
                lappend info -frame $_scenes($name-lastviewframe)
                return $info
            }
            return ""
        }
        forget {
            if {[llength $args] == 0} {
                set args $_scenes(ALL)
            }
            foreach name $args {
                set i [lsearch -exact $_scenes(ALL) $name]
                set _scenes(ALL) [lreplace $_scenes(ALL) $i $i]
                catch {unset _scenes($name-files)}
                catch {unset _scenes($name-vmdscript)}
                catch {unset _scenes($name-vmdinfo)}
                catch {unset _scenes($name-frames)}
                catch {unset _scenes($name-play)}
                catch {unset _scenes($name-show)}
                catch {unset _scenes($name-initialview)}
                catch {unset _scenes($name-lastview)}
                catch {unset _scenes($name-initialviewframe)}
                catch {unset _scenes($name-lastviewframe)}
                catch {unset _scenes($name-viewscript)}

                itcl::delete object $_scenes($name-views)
                catch {unset _scenes($name-views)}
            }
            $_dispatcher event -idle !rebuild
        }
        default {
            error "bad option \"$option\": should be define, show, vmdinfo, initialview, lastview, clear, forget"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: getimages frame <num> <frameCallback> <doneCallback>
# USAGE: getimages anim <spec> <frameCallback> <doneCallback>
#
# Used to query an image from the current view.  The "frame" option
# requests a single frame.  The "anim" option requests a series of
# frames for an animation.  Calls the given <frameCallback> command
# once for each frame of the image with the frame number and image
# handle as arguments.  Calls the <doneCallback> at the end of all
# frames to finalize the operation.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::getimages {what spec framecb donecb} {
    if {$_getimages(remaining) > 0} {
        error "another download operation is already in progress"
    }

    switch -- $what {
        frame {
            if {$spec eq "current"} {
                set spec $_play(frame)
            } elseif {![string is integer -strict $spec] || $spec < 0} {
                error "bad value \"$spec\": should be integer >= 0"
            }

            if {[$_cache get $spec imh]} {
                after idle $framecb $spec $imh
                after idle $donecb
            } else {
                set _getimages(playinfo) [array get _play]
                set _getimages(remaining) 1
                set _getimages(framecallback) $framecb
                set _getimages(donecallback) $donecb

                set _play(frame) $spec
                RefreshFrame
            }
        }
        anim {
            error "not yet implemented"
        }
        default {
            error "bad option \"$what\": should be frame or anim"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: reset ?-initialview|-lastview?
#
# Resets the rotation/pan/zoom for the current scene.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::reset {{fromView "-initialview"}} {
    # if we're still playing, we should stop now
    if {$_play(state) ne "pause"} {
        play pause
    }

    if {$fromView eq "-initialview"} {
        set _defview yes  ;# play through pre-defined animation views
    } else {
        set _defview no
    }

    # unpack the initial settings into _camera so we can use them later
    if {$_scurr ne ""} {
        set val [scene vmdinfo $_scurr rendermode]
        if {$val eq "GLSL"} {
            set _camera(quality) "high"
        } else {
            set _camera(quality) "normal"
        }
        SendBytes "setquality $_camera(quality)\n"

        foreach {key mtx} $_scenes($_scurr$fromView) {
            switch -- $key {
                -scale {
                    set _camera(scale) [lindex [lindex $mtx 0] 0]
                }
                -center {
                    set _camera(centerx) [lindex [lindex $mtx 0] 3]
                    set _camera(centery) [lindex [lindex $mtx 1] 3]
                    set _camera(centerz) [lindex [lindex $mtx 2] 3]
                }
            }
        }
        SendBytes "setview $_scenes($_scurr$fromView)\n"
    }

    # see if there's a starting frame from -play or -frames,
    # or from the last view of the scene
    set startfrom ""
    if {$_scurr ne ""} {
        if {$_defview} {
            set startfrom $_scenes($_scurr-initialviewframe)
        } else {
            set startfrom $_scenes($_scurr${fromView}frame)
        }
    }
    if {$startfrom eq ""} {
        set startfrom $_play(startfrom)
    }

    if {$startfrom ne ""} {
        # does this scene have a starting frame? jump there
        InvokeChangeCommand at $startfrom
        set _play(frame) $startfrom
    }

    # tell the server to reset
    if {[IsConnected]} {
        NewEpoch
        RefreshFrame
    }
}

# ----------------------------------------------------------------------
# USAGE: mouseMode ?<newMode>?
#
# Clients use this to set the mode for the mouse controls -- one of
# "rotate", "pan", or "zoom".  If a mode isn't specified, then the
# current mode is returned.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::mouseMode {{newMode ""}} {
    if {$newMode eq ""} {
        return $_mode
    }

    set allowedModes {rotate pan zoom}
    if {[lsearch $allowedModes $newMode] < 0} {
        error "bad mode \"$newMode\": should be one of [join $allowedModes {, }]"
    }

    # remove the current mode from the viewer widget
    set btags [bindtags $itk_component(3dview)]
    if {$_mode ne ""} {
        set i [lsearch $btags $_mode-$this]
        if {$i >= 0} {
            set btags [lreplace $btags $i $i]
        }
    }

    # add the tag for this mode onto the widget
    set btags [linsert $btags 0 $newMode-$this]
    bindtags $itk_component(3dview) $btags

    set _mode $newMode
}

# ----------------------------------------------------------------------
# USAGE: play play>>
# USAGE: play <<play
# USAGE: play pause
# USAGE: play step>
# USAGE: play <step
# USAGE: play goto <frameNum>
# USAGE: play mode once|loop|rock
# USAGE: play speed ?<factor>?
# USAGE: play smooth ?<factor>?
#
# Clients use this to control the playing of animation frames.  The
# "play>>" and "<<play" options cause frame to play forward/backward
# at the speed set by "play speed".  The smooth factor can be set to
# 0 for no smoothing, or to a positive integer value for smoothing over
# that may frames before/after the current one.
#
# The frames may play once to the end/beginning, and then loop or rock
# back through depending on "play mode".  The "pause" operation stops
# on the current frame.  The "step>" and "<step" options step one frame
# forward or backward.  The "goto" option seeks a particular frame and
# pauses there.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::play {option args} {
    switch -- $option {
        play>> {
            set _play(state) "play>>"

            # set the range of frames for loop/rock modes
            set _play(frame0) 0
            set _play(frame1) [expr {$_play(maxframes)-1}]
            if {$_play(range) ne ""} {
                foreach {r0 r1} $_play(range) break
                if {$_play(frame) >= $r0 && $_play(frame) <= $r1} {
                    set _play(frame0) $r0
                    set _play(frame1) $r1
                }
            }

            RequestFrames
            set _play(afterId) [after 1 \
                [itcl::code $this NextFrame $_play(frame)]]
        }
        <<play {
            set _play(state) "<<play"

            # set the range of frames for loop/rock modes
            set _play(frame0) 0
            set _play(frame1) [expr {$_play(maxframes)-1}]
            if {$_play(range) ne ""} {
                foreach {r0 r1} $_play(range) break
                if {$_play(frame) >= $r0 && $_play(frame) <= $r1} {
                    set _play(frame0) $r0
                    set _play(frame1) $r1
                }
            }

            RequestFrames
            set _play(afterId) [after 1 \
                [itcl::code $this NextFrame $_play(frame)]]
        }
        pause {
            if {$_play(state) eq "spinning"} {
                set num $_play(rotate)
                if {[info exists _autospin($num)]} {
                    # send the orientation for this final frame
                    SendBytes "[list setview -rotate $_autospin($num)]\n"
                }
                RefreshFrame
            }

            set _play(state) "pause"
            if {$_play(afterId) ne ""} {
                after cancel $_play(afterId)
                set _play(afterId) ""
            }
            InvokeChangeCommand pause
            return
        }
        step> {
            play goto [expr {$_play(frame)+1}]
        }
        <step {
            play goto [expr {$_play(frame)-1}]
        }
        goto {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"play goto num\""
            }
            set num [lindex $args 0]
            if {![string is integer -strict $num]} {
                error "bad value \"$num\": should be positive integer"
            }
            if {$num < 0} {
                set num 0
            }
            if {$num >= $_play(maxframes)} {
                set num [expr {$_play(maxframes)-1}]
            }

            # seeking a frame stops any play
            if {$_play(state) ne "pause"} {
                play pause
            }

            # if we're seeking something we don't have, then cancel any
            # frames that we've requested previously
            if {![$_cache exists $num]} {
                NewEpoch
            }
            ShowFrame $num
            InvokeChangeCommand goto $num
        }
        mode {
            if {[llength $args] == 0} {
                return $_play(mode)
            }
            set val [lindex $args 0]
            set allowed {once loop rock}
            if {[lsearch $allowed $val] < 0} {
                error "bad value \"$val\": should be one of [join $allowed {, }]"
            }
            set _play(mode) $val
        }
        speed {
            if {[llength $args] == 0} {
                return $_play(speed)
            }
            set val [lindex $args 0]
            if {![string is double -strict $val]} {
                error "bad value \"$val\": should be number between -1 and 1"
            }

            set maxd 1000
            set midd 250
            set mind 10
            if {$val < -1} {
                set _play(delay) $maxd
            } elseif {$val < 0} {
                set _play(delay) [expr {round(-$val*($maxd-$midd)+$midd)}]
            } elseif {$val < 1} {
                set _play(delay) [expr {round((1-$val)*($midd-$mind)+$mind)}]
            } else {
                set _play(delay) $mind
            }
            set _play(speed) $val
        }
        smooth {
            if {[llength $args] == 0} {
                return $_play(smooth)
            }
            set val [lindex $args 0]
            if {![string is double -strict $val]} {
                error "bad value \"$val\": should be number between -1 and 1"
            }
            if {$val <= -1} {
                set val 0
            } elseif {$val <= 0} {
                # map (-1,0) to values 0-3
                set val [expr {round((1+$val)*3)}]
            } elseif {$val < 1} {
                # map (0,1) to values 3-10
                set val [expr {round($val*7 + 3)}]
            } else {
                set val 10
            }

            if {$val != $_play(smooth)} {
                set _play(smooth) $val
                $_cache clear
                NewEpoch
                SendBytes "smoothreps $val\n"
            }
        }
        default {
            error "bad option \"$option\": should be one of >>, <<, step>, <step, pause, goto, mode, speed, smooth"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: waiting ?<state>?
#
# Used to query or change the "waiting" status of this widget.  When
# the "waiting" mode is turned on, a pop-up message appears and the
# -waitingcommand is invoked with the current status to notify the
# main program.  When the "waiting" mode is turned off, the message
# is taken down and the -waitingcommand is invoked again.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::waiting {{state ""}} {
    if {$state eq ""} {
        return $_waiting
    }
    if {![string is boolean -strict $state]} {
        error "bad value \"$state\": should be boolean"
    }

    if {$state != $_waiting} {
        set _waiting $state
        if {$state} {
            set x [winfo rootx $itk_component(3dview)]
            set y [winfo rooty $itk_component(3dview)]
            set w [winfo width $itk_component(3dview)]
            set h [winfo height $itk_component(3dview)]

            set mesg $itk_component(hull).loading
            set mesgw [winfo reqwidth $mesg]
            set mesgh [winfo reqheight $mesg]

            set mx [expr {$x + ($w-$mesgw)/2}]
            set my [expr {$y + ($h-$mesgh)/2}]
            wm geometry $mesg +$mx+$my

            # delay a little so it doesn't show for short load times
            # note: can't do "after 250 wm deiconify" -- that messes up
            # communication with the server.  Use raise/lower instead.
            lower $mesg
            wm deiconify $mesg
            if {$w > 1 || $h > 1} {
                after 250 "raise $mesg"
            }

            AnimateWaiting start
        } else {
            AnimateWaiting stop
            wm withdraw $itk_component(hull).loading
        }

        if {$itk_option(-waitingcommand) ne ""} {
            uplevel #0 $itk_option(-waitingcommand) $_waiting
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: spinning ?<state>?
#
# Used to query or change the "spinning" status of this widget.  This
# is like the "waiting" mode, but with a tiny spinner in the corner.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::spinning {{state ""}} {
    if {$state eq ""} {
        return $_spinning
    }
    if {![string is boolean -strict $state]} {
        error "bad value \"$state\": should be boolean"
    }

    if {$state != $_spinning} {
        set _spinning $state
        if {$state} {
            set spinner $itk_component(hull).play
            set x [winfo rootx $itk_component(3dview)]
            set w [winfo width $itk_component(3dview)]
            set y [winfo rooty $itk_component(3dview)]
            set sw [winfo reqwidth $spinner]

            wm geometry $spinner +[expr {$x+$w-$sw}]+$y
            wm deiconify $spinner
            raise $spinner
            AnimateWaiting start
        } else {
            AnimateWaiting stop
            wm withdraw $itk_component(hull).play
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: quality ?high|normal?
#
# Used to set or query the rendering quality of the scene.  This
# translates to the VMD rendermode GLSL or Normal.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::quality {{newval ""}} {
    if {$newval ne ""} {
        if {[lsearch {normal high} $newval] < 0} {
            error "bad quality value \"$newval\": should be normal or high"
        }
        if {$newval ne $_camera(quality)} {
            set _camera(quality) $newval
            SendBytes "setquality $_camera(quality)\n"
        }
    }
    return $_camera(quality)
}

# ----------------------------------------------------------------------
# USAGE: Connect
#
# Establishes a connection to a new visualization server.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::Connect {} {
    if { [IsConnected] } {
        return 1
    }
    set hosts [GetServerList "vmdmds"]
    if { $hosts eq "" } {
        return 0
    }
    set result [VisViewer::Connect $hosts]
    if { $result } {
        $_dispatcher event -idle !rebuild
    }
    return $result
}

# ----------------------------------------------------------------------
# USAGE: Disconnect
#
# Clients use this method to disconnect from the current rendering
# server.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::Disconnect {} {
    VisViewer::Disconnect

    play pause
    set _data(loaded) ""
}

# ----------------------------------------------------------------------
# Usage ReceiveImage epoch <num> frame <num> length <size>
#
# Invoked automatically whenever the "image" command comes in from
# the rendering server.  Indicates that binary image data with the
# specified <size> will follow.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::ReceiveImage {args} {
    # we got a frame -- must be done loading
    waiting off

    # if we're not waiting for anything else, then take down small spinner too
    if {[$_cache waiting] <= 1} {
        spinning off
    }

    array set opts $args
    if {![info exists opts(epoch)] || ![info exists opts(frame)]
          || ![info exists opts(length)]} {
        ServerError "bad image response from server: missing details"
        return
    }
    set data [ReceiveBytes $opts(length)]

    if {$opts(epoch) < $_epoch} {
        # frame from an old epoch or no epoch -- ignore it
        return
    }

    if {$_data(loaded) ne ""} {
        # load the data for this image into the frame number
        set num $opts(frame)
        set imh [$_cache store $opts(frame) $data]

        # is this a "getimages" query? then send back image via callback
        if {$_getimages(remaining) > 0} {
            after idle $_getimages(framecallback) $opts(frame) $imh

            if {[incr _getimages(remaining) -1] <= 0} {
                after idle $_getimages(donecallback)
            }
            return
        }

        # is this the current frame? then make sure that it's displayed
        if {$_play(state) ne "spinning"
              && ($_play(frame) ne "" && $num == $_play(frame))} {
            $_image(plot) configure \
                -width [image width $imh] -height [image height $imh]
            $_image(plot) copy $imh
        }
    } else {
        # server has no files -- show blank frame
        set w [winfo width $itk_component(3dview)]
        set h [winfo height $itk_component(3dview)]
        $_image(plot) configure -width $w -height $h
        $_image(plot) put #000 -to 0 0 $w $h
    }
}

# ----------------------------------------------------------------------
# Usage ServerError <message>
#
# Invoked automatically whenever the "oops" command comes in from
# the rendering server.  This means that something went wrong with a
# client request.  Reports the problem so we can deal with it.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::ServerError {mesg} {
    waiting off
    tk_messageBox -title "VMD Error" -icon error -message "VMD ERROR: $mesg"
}

# ----------------------------------------------------------------------
# USAGE: Rebuild
#
# Called automatically whenever something changes that affects the
# data in the widget.  Clears any existing data and rebuilds the
# widget to display new data.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::Rebuild {} {
    if {$_data(files) ne "" && $_data(files) ne $_data(loaded)} {
        SendBytes "load $_data(files)\n"
        set _data(loaded) $_data(files)

        # resend all scene definitions
        SendBytes "scene forget\n"
        foreach name $_scenes(ALL) {
            SendBytes "scene define [list $name $_scenes($name-vmdscript)]\n"
        }
        if {$_scurr ne ""} {
            SendBytes "scene show [list $_scurr] -after {saveview initialview %v}\n"
        }

        reset $_scenes($_scurr-show)
        NewEpoch

        # figure out the number of frames
        set _play(maxframes) 0
        SendBytes "tellme {maxframes %v} frames max\n"
    } else {
        set w [winfo width $itk_component(3dview)]
        set h [winfo height $itk_component(3dview)]
        $_image(plot) configure -width $w -height $h
        $_image(plot) put #000 -to 0 0 $w $h
    }
}

# ----------------------------------------------------------------------
# USAGE: BuildViews <name>
#
# Scans through all frames in a trajectory and interprets the
# -vmdscript and -viewscript to build a view matrix for each frame.
# This is done only once for each view after the max number of frames
# is known.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::BuildViews {name} {
    if {$_play(maxframes) > 0 && $_scenes($name-initialview) ne ""} {
        set views $_scenes($name-views)
        if {[$views num] == 0} {
            # apply the initial view that we get after the "view" script
            foreach {key mtx} $_scenes($name-initialview) {
                switch -- $key {
                    -rotate { $scriptvals(rotatemtx) put $mtx }
                    -scale  { $scriptvals(scalemtx) put $mtx }
                    -center { $scriptvals(centermtx) put $mtx }
                    -global { $scriptvals(globalmtx) put $mtx }
                }
            }

            # set the $mol() var used in the view script
            $scriptparser load [list catch {unset mol}]
            $scriptparser load [list set mol(top) 0]
            foreach fname $_scenes($name-files) {
                if {[regexp {^@name:(.+)} $fname match id]} {
                    $scriptparser load [list set mol($id) 0]
                }
            }

            # run through each frame and see what the view script does to it
            set script $_scenes($_scurr-viewscript)

            for {set n 0} {$n < $_play(maxframes)} {incr n} {
                # catch up on any outstanding animation requests
                set newops ""
                foreach rec $scriptvals(animateops) {
                    foreach {num cmd} $rec break
                    ScanScript $cmd -continue
                    if {[incr num -1] > 0} {
                        lappend newops [list $num $cmd]
                    }
                }
                set scriptvals(animateops) $newops

                # set the $frame var used in the view script
                $scriptparser load [list set frame $n]

                # execute the script and see what values come out
                if {[catch {ScanScript $script -continue} result]} {
                    ServerError "Error in view script at frame $n: $result"
                    break
                }
                catch {unset vals}
                array set vals $result

                $views setview $n \
                    -rotate [$vals(rotatemtx) get] \
                    -scale [$vals(scalemtx) get] \
                    -center [$vals(centermtx) get] \
                    -global [$vals(globalmtx) get]
            }

            # now that we've rebuilt the views, we should refresh the view
            RefreshFrame
        }

        # send views for all frames back to the server
        set mnames {rotate_matrix scale_matrix center_matrix global_matrix}
        for {set n 0} {$n < $_play(maxframes)} {incr n} {
            catch {unset mtx}
            array set mtx [$views getview $n]
            set mvals [list $mtx(-rotate) $mtx(-scale) $mtx(-center) $mtx(-global)]
            SendBytes "frames defview $n [list $mnames $mvals]\n"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: DoResize
#
# Handles the !resize event in response to the widget changing size.
# Sends the new width/height to the server.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::DoResize {} {
    NewEpoch

    set w [winfo width $itk_component(3dview)]
    set h [winfo height $itk_component(3dview)]
    set _screenSize [expr {0.5*($w+$h)}]

    # save the current screen size in case server produces smaller images
    # during drag operations to go faster
    $_cache configure -iwidth $w -iheight $h

    SendBytes "resize $w $h\n"
    RefreshFrame
}

# ----------------------------------------------------------------------
# USAGE: Mouse click x y func
#        Mouse drag x y func
#        Mouse release x y func
#
# Called automatically when the user clicks and drags in one of the
# various modes (rotate, pan, zoom).  Computes the change since the
# last move and then invokes <func> to perform the view function.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::Mouse {option x y func} {
    switch -- $option {
        click {
            play pause
            NewEpoch
            SendBytes "drag start\n"
            RefreshFrame

            set _mevent(x0) $x
            set _mevent(y0) $y
            set _mevent(xlast) $x
            set _mevent(ylast) $y
            set _mevent(dx) 0
            set _mevent(dy) 0

            set _mevent(time) [clock clicks -milliseconds]
            set _mevent(velocities) ""

            set _mevent(zoom) $_camera(zoom)
            set _mevent(rotations) ""

            $itk_component(3dview) configure -cursor $func2cursor($func)
        }
        drag {
            # compute mouse movement in pixels
            set _mevent(dx) [set dx [expr {$x-$_mevent(xlast)}]]
            set _mevent(dy) [set dy [expr {$y-$_mevent(ylast)}]]

            # compute time delay (for acceleration)
            set time [clock clicks -milliseconds]
            if {$time < $_mevent(time)} {
                set _mevent(time) [expr {4.0*(1<<30)+$_mevent(time)}]
            }
            set dt [expr {$time-$_mevent(time)}]
            if {$dt < 1} {set dt 1}
            set _mevent(time) $time

            # compute rolling velocity for movements
            set vel [expr {sqrt($dx*$dx+$dy*$dy)/$dt}]
            set _mevent(velocities) [lrange [linsert $_mevent(velocities) 0 $time $vel] 0 19]

            $func mouse $x $y
            RefreshFrame

            set _mevent(xlast) $x
            set _mevent(ylast) $y
        }
        release {
            Mouse drag $x $y $func

            set cutoff [expr {$_mevent(time)-200}]
            set avgvel 0
            set npts 0
            foreach {time vel} $_mevent(velocities) {
                if {$time < $cutoff} {
                    break
                }
                set avgvel [expr {$avgvel+$vel}]
                incr npts
            }
            if {$npts > 0} {
                set avgvel [expr {$avgvel/double($npts)}]
            }

            SendBytes "drag end\n"
            if {$func eq "rotate" && $avgvel > 0.001} {
                # compute the average rotation angles over the past few ops
                set avg(mx) 0
                set avg(my) 0
                set avg(mz) 0
                set avg(npts) 0
                foreach {mx my mz} $_mevent(rotations) {
                    foreach comp {mx my mz} {
                        set avg($comp) [expr {$avg($comp)+[set $comp]}]
                    }
                    incr avg(npts)
                }
                foreach comp {mx my mz} {
                    set avg($comp) [expr {$avg($comp)/double($avg(npts))}]
                }

                if {$avg(mx) != 0 || $avg(my) != 0 || $avg(mz) != 0} {
                    # emulate VMD behavior that keeps the molecule spinning
                    catch {unset _autospin}

                    if {$avgvel > 0.2} {
                        set nframes 36
                    } else {
                        set nframes [expr {round(36+180*(0.2-$avgvel))}]
                    }

                    NewEpoch
                    set _play(state) "spinning"
                    SendBytes "frames rotate $_epoch $avg(mx) $avg(my) $avg(mz) $nframes\n"
                    NextRotate 0 $nframes

                    # put up the spinner while we're waiting for frames
                    for {set i 0} {$i < $nframes} {incr i} {
                        $_cache waiting set $i
                    }
                    spinning on
                }
            }
            $itk_component(3dview) configure -cursor ""
        }
        default {
            error "bad option \"$option\": should be click, drag, release"
        }
    }
}
    
# ----------------------------------------------------------------------
# USAGE: rotate mouse x y
# USAGE: rotate x <degrees>
# USAGE: rotate y <degrees>
# USAGE: rotate z <degrees>
#
# Called automatically when the user clicks and drags while in "rotate"
# mode, or when a client requests rotation around x, y, or z.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::rotate {option args} {
    switch -- $option {
        "mouse" {
            set x [lindex $args 0]
            set y [lindex $args 1]

            set w [winfo width $itk_component(3dview)]
            set h [winfo height $itk_component(3dview)]
            set x1 [expr {double($w) / 3}]
            set x2 [expr {$x1 * 2}]
            set y1 [expr {double($h) / 3}]
            set y2 [expr {$y1 * 2}]
            set dx [expr {$x - $_mevent(x0)}]
            set dy [expr {$y - $_mevent(y0)}]
            set mx 0
            set my 0
            set mz 0

            if { $_mevent(x0) < $x1 } {
                set mz $dy
            } elseif { $_mevent(x0) < $x2 } {
                set mx $dy
            } else {
                set mz [expr {-$dy}]
            }

            if { $_mevent(y0) < $y1 } {
                set mz [expr {-$dx}]
            } elseif { $_mevent(y0) < $y2 } {
                set my $dx
            } else {
                set mz $dx
            }
            SendBytes "rotate x by $mx; rotate y by $my; rotate z by $mz\n"

            set _defview no

            # keep the last few rotation values
            # these come in handy when computing auto-spin
            set _mevent(rotations) [lrange [linsert $_mevent(rotations) 0 $mx $my $mz] 0 8]

            set _mevent(x0) $x
            set _mevent(y0) $y
        }
        x - y - z {
            if {$_play(state) eq "pause"} {
                set deg [lindex $args 0]
                SendBytes "rotate $option by $deg\n"
                set _defview no
                RefreshFrame
            }
        }
        default {
            error "bad option \"$option\": should be mouse, x, y, z"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: pan mouse x y
# USAGE: pan by x y
# USAGE: pan set x y
#
# Called automatically when the user clicks and drags while in "pan"
# mode, or when the user presses arrow keys to pan around.  The
# "mouse" operation represents mouse coordinates since the last
# mouse click operation.  The "by" operation represents a drag
# operation to the x,y screen coordinate.  Computes the offset for
# this pan operation and sends a command off to the server to pan
# the image.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::pan {option x y} {
    switch -- $option {
        "mouse" {
            # treat x/y as mouse coordinates relative to click point
            set dx [expr {100*($x-$_mevent(x0))/$_screenSize*$_camera(scale)}]
            set dy [expr {-100*($y-$_mevent(y0))/$_screenSize*$_camera(scale)}]
            set _mevent(x0) $x
            set _mevent(y0) $y
            SendBytes "translate by $dx $dy 0\n"

            set _defview no
        }
        "by" {
            if {$_play(state) eq "pause"} {
                # treat x/y as multiples of the scale unit
                set dx [expr {$x * $_camera(scale)}]
                set dy [expr {-($y * $_camera(scale))}]
                SendBytes "translate by $dx $dy 0\n"
                set _defview no
                RefreshFrame
            }
        }
        "set" {
            if {$_play(state) eq "pause"} {
                # treat x/y as a specific coordinate
                SendBytes "translate to $x $y 0\n"
                set _defview no
                RefreshFrame
            }
        }
        default {
            error "bad option \"$option\": should be mouse, by"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: zoom in <factor>
# USAGE: zoom out <factor>
# USAGE: zoom set <factor>
# USAGE: zoom mouse <x> <y>
#
# Called automatically when the user clicks on one of the zoom
# controls for this widget.  Changes the zoom for the current scene.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::zoom {option args} {
    switch -- $option {
        "in" {
            set factor [lindex $args 0]
            if {$factor > 0 && $_play(state) eq "pause"} {
                NewEpoch
                set _camera(zoom) [expr {$_camera(zoom) * $factor}]
                set factor [expr {$_camera(zoom) * $_camera(scale)}]
                SendBytes "vmd_scale to $factor\n"
                set _defview no
                RefreshFrame
            }
        }
        "out" {
            set factor [lindex $args 0]
            if {$factor > 0 && $_play(state) eq "pause"} {
                NewEpoch
                set _camera(zoom) [expr $_camera(zoom) / double($factor)]
                set factor [expr {$_camera(zoom) * $_camera(scale)}]
                SendBytes "vmd_scale to $factor\n"
                set _defview no
                RefreshFrame
            }
        }
        "set" {
            set factor [lindex $args 0]
            if {$factor > 0 && $_play(state) eq "pause"} {
                NewEpoch
                set _camera(zoom) $factor
                set factor [expr {$_camera(zoom) * $_camera(scale)}]
                SendBytes "vmd_scale to $factor\n"
                set _defview no
                RefreshFrame
            }
        }
        "mouse" {
            set x [lindex $args 0]
            set y [lindex $args 1]
            set dx [expr {$x-$_mevent(x0)}]
            set dy [expr {$y-$_mevent(y0)}]
            set len [expr {sqrt($dx*$dx + $dy*$dy)}]

            set factor [expr {abs($_screenSize-$len)/double($_screenSize)}]

            if {$dy > 0} {
                # mouse moving down -- reduce (zoom out)
                # scale down the original zoom factor captured on mouse click
                set _camera(zoom) [expr {$_mevent(zoom)*$factor}]
                set factor [expr {$_camera(zoom) * $_camera(scale)}]
                SendBytes "vmd_scale to $factor\n"
            } else {
                # mouse moving up -- magnify (zoom in)
                # scale up the original zoom factor captured on mouse click
                set _camera(zoom) [expr {$_mevent(zoom)/double($factor)}]
                set factor [expr {$_camera(zoom) * $_camera(scale)}]
                SendBytes "vmd_scale to $factor\n"
            }
            set _defview no
        }
        default {
            error "bad option \"$option\": should be mouse, in, out"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: NewEpoch
#
# Used to clear any animation frames and bump the epoch counter whenever
# the view changes and the animation is outdated.  Subsequent calls to
# RequestFrames will get frames for this new epoch from the server.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::NewEpoch {} {
    $_cache clear
    incr _epoch
}

# ----------------------------------------------------------------------
# USAGE: MaxFrames ?max?
#
# Used to set the maximum number of frames in an animation.  Usually
# sent back from the server via a "tellme" command.  Can also be used
# to query the maximum number of frames if the number isn't specified.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::MaxFrames {{max ""}} {
    if {$max eq ""} {
        return $_play(maxframes)
    }
    set _play(maxframes) $max
    if { $max > 0 } {
	InvokeChangeCommand maxframes $max
    }

    # if we haven't figured out the view matrices yet, do it now
    if {$_scurr ne ""} {
        BuildViews $_scurr
    }

    # jump to the first frame for the current scene
    if {$_scurr ne "" && $_play(startfrom) ne ""} {
        if {$_play(startfrom) ne ""} {
            # does this scene have a starting frame? jump there
            InvokeChangeCommand at $_play(startfrom)
            set _play(frame) $_play(startfrom)
        }

        # if the -play option is set, start playing
        if {[lsearch {once loop rock} $_play(startop)] >= 0} {
            InvokeChangeCommand playmode $_play(startop) $_play(startdir)
        }
    } else {
        # adjust the current frame if it falls out of the new bounds
        if {$max == 0} {
            BlankFrame
        } elseif {$_play(frame) >= $_play(maxframes)} {
            set _play(frame) [expr {$_play(maxframes)-1}]
        }
    }

    # this is a good spot to send the smoothing factor, since mol is loaded
    SendBytes "smoothreps $_play(smooth)\n"
    RefreshFrame
}

# ----------------------------------------------------------------------
# USAGE: SaveView ?<what>? ?<viewName>? ?-rotate m -scale m -center m -global m?
#
# Used to get/set the initial view computed by VMD when a molecule
# is loaded.  This info is sent back by the server, and this method
# stores the info for later.  The <what> value can be "initialview"
# or "lastview", depending on what view we want to store.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::SaveView {args} {
    if {[llength $args] == 0} { return "" }

    set what [lindex $args 0]
    set name [lindex $args 1]
    if {[info exists _scenes($name-$what)]} {
        if {[llength $args] == 2} {
            # just querying current values
            return $_scenes($name-$what)
        }

        # new values given -- set values
        set _scenes($name-$what) [lrange $args 2 end]

        if {$what eq "initialview"} {
            # just got initial view -- use it to set view parameters
            foreach {key mtx} $_scenes($name-$what) {
                switch -- $key {
                    -scale {
                        set _camera(scale) [lindex [lindex $mtx 0] 0]
                    }
                    -center {
                        set _camera(centerx) [lindex [lindex $mtx 0] 3]
                        set _camera(centery) [lindex [lindex $mtx 1] 3]
                        set _camera(centerz) [lindex [lindex $mtx 2] 3]
                    }
                }
            }

            # if we haven't figured out the view matrices yet, do it now
            BuildViews $name
        }
    }
    return ""
}

# ----------------------------------------------------------------------
# USAGE: BlankFrame
#
# Clears out the frame to a blank background.  Used whenever the
# view changes to clear things before the new image comes back from
# the server.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::BlankFrame {} {
    set bg #000
    if {$_scurr ne ""} {
        set bggrad [scene vmdinfo $_scurr backgroundgradient]
        if {!$bggrad} {
            set bg [scene vmdinfo $_scurr color-Background]

            array set VMDcolors [scene vmdinfo $_scurr colorvals]
            if {[info exists VMDcolors($bg)]} {
                # if we have the exact color definition from VMD, then use that.
                # VMD's defintion of "gray" and "blue2" is a little different
                # from the usual X convention.
                set bg $VMDcolors($bg)
            }
        }
    }

    set w [winfo width $itk_component(3dview)]
    set h [winfo height $itk_component(3dview)]
    $_image(plot) put $bg -to 0 0 $w $h
}

# ----------------------------------------------------------------------
# USAGE: RefreshFrame
#
# Sprinkled in at various places where the view parameters change
# to request a frame update.  If an animation is in progress, this
# does nothing.  Otherwise, it causes the server to send a new copy
# of the current frame.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::RefreshFrame {} {
    if {$_play(state) eq "pause" && $_play(maxframes) > 0} {
        set dflag [expr {($_defview) ? "-defview" : ""}]
        SendBytes "frames time $_epoch $_play(frame) $dflag\n"
    }
}

# ----------------------------------------------------------------------
# USAGE: RequestFrames
#
# Used to request frames from the server for playing an animation
# forward or backward.  This is called when the animation first starts,
# and should be called as each frame is played to see if we have enough
# frames in the cache.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::RequestFrames {} {
    if {$_play(maxframes) == 0} {
        # animation is not possible -- bail out!
        return
    }

    # look ahead and see if we should request frames from the server
    set dflag [expr {($_defview) ? "-defview" : ""}]

    foreach {op val} [$_cache lookAhead $_play(frame) $_play(state) $_play(mode) $_play(frame0) $_play(frame1) $_play(frameby)] {
        switch -- $op {
            frame {
                SendBytes "frames time $_epoch $val $dflag\n"
            }
            range {
                foreach {v0 step v1} [split $val :] break
                SendBytes "frames time $_epoch $v0 $v1 $step $dflag\n"

                # put up the spinner while we're waiting for frames
                spinning on
            }
            default {
                error "oops! bad look ahead operation \"$op\""
            }
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: NextFrame <num>
#
# Used internally to advance the animation when playing forward or
# backward.  The <num> is the next frame that should be displayed
# in the viewer.  If the animation is still playing, this call is
# set up in an after event to get called again later.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::NextFrame {num} {
    # show the given image
    if {![ShowFrame $num]} {
        # if that fails, try again later
        set _play(afterId) [after 50 [itcl::code $this NextFrame $num]]
        return
    }

    # request future frames if we'll need them soon
    RequestFrames

    # figure out the next frame to show
    switch -- $_play(state) {
        play>> {
            set nextnum [expr {$num + $_play(frameby)}]
            if {$nextnum > $_play(frame1)} {
                switch -- $_play(mode) {
                    once {
                        # all done with play
                        play pause
                        return
                    }
                    loop {
                        # start over at the beginning
                        set nextnum $_play(frame0)
                    }
                    rock {
                        # play backward from this frame
                        set _play(state) "<<play"
                        set nextnum [expr {$num - $_play(frameby)}]
                        if {$nextnum < $_play(frame0)} {
                            set nextnum $_play(frame0)
                        }
                    }
                }
            }
        }
        <<play {
            set nextnum [expr {$num - $_play(frameby)}]
            if {$nextnum < $_play(frame0)} {
                switch -- $_play(mode) {
                    once {
                        # all done with play
                        play pause
                        return
                    }
                    loop {
                        # start over at the end
                        set nextnum [expr {$_play(frame1)-1}]
                        if {$nextnum < $_play(frame0)} {
                            set nextnum $_play(frame0)
                        }
                    }
                    rock {
                        # play forward from this frame
                        set _play(state) "play>>"
                        set nextnum [expr {$num + $_play(frameby)}]
                        if {$nextnum > $_play(frame1)} {
                            set nextnum $_play(frame1)
                        }
                    }
                }
            }
        }
        pause { 
            return
        }
    }

    # show the next frame after the proper delay
    set _play(afterId) [after $_play(delay) \
        [itcl::code $this NextFrame $nextnum]]
}

# ----------------------------------------------------------------------
# USAGE: ShowFrame <num>
#
# Used to load the image for a particular frame into the view.
# Returns 1 if the image could be displayed, and 0 if we're still
# waiting for it.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::ShowFrame {num} {
    if {$_play(maxframes) > 0} {
        # show this frame when it comes in
        set _play(frame) $num

        if {[$_cache get $num imh]} {
            InvokeChangeCommand at $num
            $_image(plot) configure \
                -width [image width $imh] -height [image height $imh]
            $_image(plot) copy $imh
            update  ;# force image onto screen
            return 1
        } elseif {![$_cache waiting $num]} {
            set dflag [expr {($_defview) ? "-defview" : ""}]
            SendBytes "frames time $_epoch $num $dflag\n"
            return 0
        }
    }
    return 0
}

# ----------------------------------------------------------------------
# USAGE: NextRotate <num> <max>
#
# Used internally to advance the animation when rotating the current
# view.  The <num> is a frame number in the current sequence, and
# <max> is the current maximum.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::NextRotate {num max} {
    # guard against animation getting started twice
    if {$_play(afterId) ne ""} {
        after cancel $_play(afterId)
        set _play(afterId) ""
    }

    # show the given image
    if {![ShowRotate $num]} {
        # if that fails, try again later
        set _play(afterId) [after 50 [itcl::code $this NextRotate $num $max]]
        return
    }

    if {[incr num] >= $max} {
        set num 0
    }

    # show the next frame after the proper delay
    set _play(afterId) [after $_play(delay) [itcl::code $this NextRotate $num $max]]
}

# ----------------------------------------------------------------------
# USAGE: ShowRotate <num>
#
# Used to load the image for a particular frame of molecule rotation
# into the view.  Returns 1 if the image could be displayed, and 0 if
# we're still waiting for it.  Unlike ShowFrame, this routine doesn't
# mess with the current frame or request upcoming frames.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::ShowRotate {num} {
    if {$_play(state) eq "spinning"} {
        # show this frame when it comes in
        set _play(rotate) $num

        if {[$_cache get $num imh]} {
            $_image(plot) configure \
                -width [image width $imh] -height [image height $imh]
            $_image(plot) copy $imh
            return 1
        }
    }
    return 0
}

# ----------------------------------------------------------------------
# USAGE: AutoSpinMatrix <num> <matrix>
#
# Sent by the server after auto-spin is initiated.  Defines the rotation
# <matrix> for a given frame <num>.  We need this when the rotation
# is stopped, so we know the final orientation of the molecule.  Just
# store for now and use later when play is paused.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::AutoSpinMatrix {num matrix} {
    set _autospin($num) $matrix
}

# ----------------------------------------------------------------------
# USAGE: AnimateWaiting start
# USAGE: AnimateWaiting frame <num>
# USAGE: AnimateWaiting stop
#
# Used internally to handle animation of the spinner in the "waiting"
# box that appears when the widget is waiting.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::AnimateWaiting {op {num ""}} {
    switch -- $op {
        start {
            if {$_waitinganim ne ""} {
                after cancel $_waitinganim
                set _waitinganim ""
            }
            AnimateWaiting frame 1
        }
        frame {
            $itk_component(hull).loading.spinner configure \
                -image [Rappture::icon roller$num]
            $itk_component(hull).play.spinner configure \
                -image [Rappture::icon roller$num]

            if {[incr num] > 8} {
                set num 1
            }
            set _waitinganim [after 200 \
                [itcl::code $this AnimateWaiting frame $num]]
        }
        stop {
            if {!$_waiting && !$_spinning} {
                if {$_waitinganim ne ""} {
                    after cancel $_waitinganim
                    set _waitinganim ""
                }
            }
        }
        default {
            error "bad option \"$op\": should be start, stop, frame"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: InvokeChangeCommand <arg> <arg>...
#
# Invoked whenver the widget changes state--from playing to paused,
# or when the view jumps to a frame.  Invokes the -changecommand for
# this widget to notify the VCR control or other parts of the program
# about the change.  If there is no -changecommand set, this does
# nothing.
# ----------------------------------------------------------------------
itcl::body Rappture::VMDvisViewer::InvokeChangeCommand {args} {
    if {$itk_option(-changecommand) ne ""} {
        uplevel #0 $itk_option(-changecommand) $args
    }
}
