#!/bin/sh
# -*- mode: Tcl -*-
# ----------------------------------------------------------------------
#  TEST PROGRAM for VMD
#
#  This is a test client for the client/server configuration of VMD.
#  It allows you to type commands and see the resulting image from
#
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2013  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
set installdir [file dirname [info script]]
set libdir [file join $installdir "lib"]

package require Img
package require RapptureGUI

option add *comm.font -*-courier-medium-r-normal-*-*-120-*
option add *Menu.tearOff off

# parse command line args
set Paradigm "socket"
while {[llength $argv] > 0} {
    set opt [lindex $argv 0]
    set argv [lrange $argv 1 end]

    switch -- $opt {
        -socket { set Paradigm "socket" }
        -stdio  { set Paradigm "stdio" }
        default {
            puts stderr "bad option \"$opt\": should be -socket or -stdio"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: load_molecule
#
# Prompts the user for a PDB file, and then tells the server to load
# that molecule.
# ----------------------------------------------------------------------
proc load_molecule {} {
    global widgets

    set file [tk_getOpenFile -title "Open Molecule File"]
    if {$file ne ""} {
        client_send [list load $file]
        client_send reset
        client_send [list frames time 0 0]
    }
}

# ----------------------------------------------------------------------
# USAGE: load_script
#
# Prompts the user for a text file, and then sends the text within
# that file along to the rendering widget.
# ----------------------------------------------------------------------
proc load_script {} {
    global widgets

    set file [tk_getOpenFile -title "Open Command File"]
    if {"" != $file && [catch {
            set fid [open $file r]
            fconfigure $fid -translation binary
            set info [read $fid]
            close $fid
          }] == 0} {

        client_send $info
    }
}

# ----------------------------------------------------------------------
# USAGE: send_command
#
# Invoked automatically whenever the user enters a command and
# presses <Return>.  Sends the command along to the rendering
# widget.
# ----------------------------------------------------------------------
proc send_command {} {
    global widgets
    global last_command

    set cmd [$widgets(command) get]

    if {[string length $cmd] > 0} {
        set last_command $cmd
    } else {
        set cmd $last_command
    }

    if {[string index $cmd 0] eq "?"} {
        set cmd "tellme {report %v} [string range $cmd 1 end]"
    }
    client_send $cmd
    $widgets(command) delete 0 end
}

# ----------------------------------------------------------------------
# USAGE: reset
#
# Used internally to reset the connection to the rendering server.
# Discards all data and resets the widget connection to the server.
# ----------------------------------------------------------------------
proc reset {} {
    global widgets

    $widgets(comm) configure -state normal
    $widgets(comm) delete 1.0 end
    $widgets(comm) configure -state disabled

    client_close
    client_connect
}

# ----------------------------------------------------------------------
# USAGE: show_comm <channel> <data>
#
# Invoked automatically whenever there is communication between
# the rendering widget and the server.  Eavesdrops on the communication
# and posts the commands in a text viewer.
# ----------------------------------------------------------------------
proc show_comm {channel {data ""}} {
    global widgets

    $widgets(comm) configure -state normal
    switch -- $channel {
        closed {
            $widgets(comm) insert end "--CLOSED--\n" error
        }
        <<line {
            $widgets(comm) insert end $data incoming "\n" incoming
        }
        >>line {
            $widgets(comm) insert end $data outgoing "\n" outgoing
        }
        error {
            $widgets(comm) insert end $data error "\n" error
        }
        default {
            $widgets(comm) insert end "$data\n"
        }
    }
    $widgets(comm) configure -state disabled
    $widgets(comm) see end
}

# ----------------------------------------------------------------------
# MAIN WINDOW
# ----------------------------------------------------------------------
menu .mbar
menu .mbar.file
.mbar.file add command -label "Load molecule..." -underline 0 -command load_molecule
.mbar.file add command -label "Load script..." -underline 5 -command load_script
.mbar.file add command -label "Reset" -underline 0 -command reset
.mbar.file add separator
.mbar.file add command -label "Exit" -underline 1 -command exit
.mbar add cascade -label "File" -underline 0 -menu .mbar.file

. configure -menu .mbar

Rappture::Panes .main -sashwidth 4 -sashrelief raised -sashpadding 4 \
    -width 6i -height 6i
pack .main -expand yes -fill both

set f [.main pane 0]
.main fraction 0 0.8

image create photo SnapShot
label $f.viewer -image SnapShot
pack $f.viewer -expand yes -fill both
set widgets(viewer) $f.viewer

set f [.main insert end -fraction 0.5]
frame $f.send
pack $f.send -side bottom -fill x
label $f.send.l -text "Send:"
pack $f.send.l -side left
set widgets(command) [entry $f.send.e]
pack $f.send.e -side left -expand yes -fill x
bind $f.send.e <KeyPress-Return> send_command

scrollbar $f.sb -orient vertical -command "$f.comm yview"
pack $f.sb -side right -fill y
text $f.comm -wrap char -yscrollcommand "$f.sb set"
pack $f.comm -expand yes -fill both
set widgets(comm) $f.comm

$widgets(comm) tag configure error -foreground red \
    -font -*-courier-medium-o-normal-*-*-120-*
$widgets(comm) tag configure incoming -foreground blue

# ----------------------------------------------------------------------
# CLIENT INTERPRETER
# ----------------------------------------------------------------------
set backend ""

proc cmd_image {args} {
    global backend
    if {$backend ne ""} {
        array set opts $args
        show_comm <<line "image: $args"

        set bytes [read $backend $opts(length)]
        show_comm <<line "<read $opts(length) bytes>"

        SnapShot configure -width 0 -height 0
        SnapShot blank
        if {[catch {SnapShot configure -data $bytes} err]} {
            show_comm error "bad image data: $err"
        }
    }
}

proc cmd_scale {args} {
    tk_messageBox -icon info -message "SCALE: $args"
}

proc cmd_report {args} {
    tk_messageBox -icon info -message $args
}

proc cmd_oops {msg} {
    tk_messageBox -icon error -message $msg
}

set parser [interp create -safe]
$parser alias nv>image cmd_image
$parser alias nv>scale cmd_scale
$parser alias nv>oops cmd_oops
$parser alias nv>report cmd_report

proc client_handle {sid} {
    global backend parser buffer

    if {[gets $sid request] < 0} {
        catch {close $sid}
        set backend ""
        show_comm closed
        tk_messageBox -icon warning -message "Lost connection to server"
    } else {
        show_comm <<line $request
        append buffer $request "\n"
        if {[regexp {^nv>} $buffer]} {
            if {[info complete $buffer]} {
                set request $buffer
                set buffer ""
                if {[catch {$parser eval $request} result] != 0} {
                    show_comm error $result
                }
            }
        } else {
            set buffer ""
        }
    }
}

proc client_send {cmd} {
    global backend
    if {$backend != ""} {
        show_comm >>line $cmd
        puts $backend $cmd
    }
}

proc client_connect {} {
    global Paradigm installdir backend widgets

    switch -- $Paradigm {
        socket {
            if {[catch {socket render07.hubzero.org 2018} sid] == 0} {
                fileevent $sid readable [list client_handle $sid]
                fconfigure $sid -buffering line
                set backend $sid
                $widgets(command) configure -state normal
                show_comm <<line "version = [gets $sid]"
            } else {
                tk_messageBox -icon error -message "Couldn't connect to server"
            }
        }
        stdio {
            set serverscript [file join $installdir vmdserver.tcl]
            if {[catch {open "| vmd -startup $serverscript -args -stdio" r+} sid] == 0} {
                fileevent $sid readable [list client_handle $sid]
                fconfigure $sid -buffering line
                set backend $sid
                $widgets(command) configure -state normal
                show_comm <<line "version = [gets $sid]"
            } else {
                tk_messageBox -icon error -message "Couldn't connect to server"
            }
        }
        default {
            error "don't know how to connect via $Paradigm"
        }
    }
}

proc client_close {} {
    global widgets backend

    if {$backend ne ""} {
        catch {close $backend}
    }
    $widgets(command) delete 0 end
    $widgets(command) configure -state disabled

    SnapShot copy [Rappture::icon alert]
}

after 10 reset
